""" FileManagerBase is a base class for all the specific File Managers
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# pylint: disable=protected-access

import six
import os
import stat

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.List import intListToString
from DIRAC.Core.Utilities.Pfn import pfnunparse


class FileManagerBase(object):
    """Base class for all the specific File Managers"""

    def __init__(self, database=None):
        self.db = database
        self.statusDict = {}

    def _getConnection(self, connection):
        if connection:
            return connection
        res = self.db._getConnection()
        if res["OK"]:
            return res["Value"]
        gLogger.warn("Failed to get MySQL connection", res["Message"])
        return connection

    def setDatabase(self, database):
        self.db = database

    def getFileCounters(self, connection=False):
        """Get a number of counters to verify the sanity of the Files in the catalog"""
        connection = self._getConnection(connection)

        resultDict = {}
        req = "SELECT COUNT(*) FROM FC_Files;"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Files"] = res["Value"][0][0]

        req = "SELECT COUNT(FileID) FROM FC_Files WHERE FileID NOT IN ( SELECT FileID FROM FC_Replicas )"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Files w/o Replicas"] = res["Value"][0][0]

        req = "SELECT COUNT(RepID) FROM FC_Replicas WHERE FileID NOT IN ( SELECT FileID FROM FC_Files )"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Replicas w/o Files"] = res["Value"][0][0]

        treeTable = self.db.dtree.getTreeTable()
        req = "SELECT COUNT(FileID) FROM FC_Files WHERE DirID NOT IN ( SELECT DirID FROM %s)" % treeTable
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        resultDict["Orphan Files"] = res["Value"][0][0]

        req = "SELECT COUNT(FileID) FROM FC_Files WHERE FileID NOT IN ( SELECT FileID FROM FC_FileInfo)"
        res = self.db._query(req, connection)
        if not res["OK"]:
            resultDict["Files w/o FileInfo"] = 0
        else:
            resultDict["Files w/o FileInfo"] = res["Value"][0][0]

        req = "SELECT COUNT(FileID) FROM FC_FileInfo WHERE FileID NOT IN ( SELECT FileID FROM FC_Files)"
        res = self.db._query(req, connection)
        if not res["OK"]:
            resultDict["FileInfo w/o Files"] = 0
        else:
            resultDict["FileInfo w/o Files"] = res["Value"][0][0]

        return S_OK(resultDict)

    def getReplicaCounters(self, connection=False):
        """Get a number of counters to verify the sanity of the Replicas in the catalog"""

        connection = self._getConnection(connection)
        req = "SELECT COUNT(*) FROM FC_Replicas;"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        return S_OK({"Replicas": res["Value"][0][0]})

    ######################################################
    #
    # File write methods
    #

    def _insertFiles(self, lfns, uid, gid, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _deleteFiles(self, toPurge, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _insertReplicas(self, lfns, master=False, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _findFiles(self, lfns, metadata=["FileID"], allStatus=False, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _getFileReplicas(self, fileIDs, fields_input=["PFN"], allStatus=False, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _getFileIDFromGUID(self, guid, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def getLFNForGUID(self, guids, connection=False):
        """Returns the LFN matching a given GUID"""
        return S_ERROR("To be implemented on derived class")

    def _setFileParameter(self, fileID, paramName, paramValue, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _deleteReplicas(self, lfns, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _setReplicaStatus(self, fileID, se, status, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _setReplicaHost(self, fileID, se, newSE, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _getDirectoryFiles(self, dirID, fileNames, metadata, allStatus=False, connection=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _getDirectoryFileIDs(self, dirID, requestString=False):
        """To be implemented on derived class"""
        return S_ERROR("To be implemented on derived class")

    def _findFileIDs(self, lfns, connection=False):
        """To be implemented on derived class
        Should return following the successful/failed convention
        Successful is a dictionary with keys the lfn, and values the FileID"""

        return S_ERROR("To be implemented on derived class")

    def _getDirectoryReplicas(self, dirID, allStatus=False, connection=False):
        """To be implemented on derived class

        Should return with only one value, being a list of all the replicas (FileName,FileID,SEID,PFN)
        """

        return S_ERROR("To be implemented on derived class")

    def countFilesInDir(self, dirId):
        """Count how many files there is in a given Directory

        :param int dirID: directory id

        :returns: S_OK(value) or S_ERROR
        """
        return S_ERROR("To be implemented on derived class")

    def _getFileLFNs(self, fileIDs):
        """Get the file LFNs for a given list of file IDs"""
        stringIDs = intListToString(fileIDs)
        treeTable = self.db.dtree.getTreeTable()

        req = (
            "SELECT F.FileID, CONCAT(D.DirName,'/',F.FileName) from FC_Files as F,\
        %s as D WHERE F.FileID IN ( %s ) AND F.DirID=D.DirID"
            % (treeTable, stringIDs)
        )
        result = self.db._query(req)
        if not result["OK"]:
            return result

        fileNameDict = {}
        for row in result["Value"]:
            fileNameDict[row[0]] = row[1]

        failed = {}
        successful = fileNameDict
        if len(fileNameDict) != len(fileIDs):
            for id_ in fileIDs:
                if id_ not in fileNameDict:
                    failed[id_] = "File ID not found"

        return S_OK({"Successful": successful, "Failed": failed})

    def addFile(self, lfns, credDict, connection=False):
        """Add files to the catalog

        :param dict lfns: dict{ lfn : info}. 'info' is a dict containing PFN, SE, Size and Checksum
                      the SE parameter can be a list if we have several replicas to register


        """
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, info in list(lfns.items()):
            res = self._checkInfo(info, ["PFN", "SE", "Size", "Checksum"])
            if not res["OK"]:
                failed[lfn] = res["Message"]
                lfns.pop(lfn)
        res = self._addFiles(lfns, credDict, connection=connection)
        if not res["OK"]:
            for lfn in lfns.keys():
                failed[lfn] = res["Message"]
        else:
            failed.update(res["Value"]["Failed"])
            successful.update(res["Value"]["Successful"])
        return S_OK({"Successful": successful, "Failed": failed})

    def _addFiles(self, lfns, credDict, connection=False):
        """Main file adding method"""
        connection = self._getConnection(connection)
        successful = {}
        result = self.db.ugManager.getUserAndGroupID(credDict)
        if not result["OK"]:
            return result
        uid, gid = result["Value"]

        # prepare lfns with master replicas - the first in the list or a unique replica
        masterLfns = {}
        extraLfns = {}
        for lfn in lfns:
            masterLfns[lfn] = dict(lfns[lfn])
            if isinstance(lfns[lfn].get("SE"), list):
                masterLfns[lfn]["SE"] = lfns[lfn]["SE"][0]
                if len(lfns[lfn]["SE"]) > 1:
                    extraLfns[lfn] = dict(lfns[lfn])
                    extraLfns[lfn]["SE"] = lfns[lfn]["SE"][1:]

        # Check whether the supplied files have been registered already
        res = self._getExistingMetadata(list(masterLfns), connection=connection)
        if not res["OK"]:
            return res
        existingMetadata, failed = res["Value"]
        if existingMetadata:
            success, fail = self._checkExistingMetadata(existingMetadata, masterLfns)
            successful.update(success)
            failed.update(fail)
            for lfn in list(success) + list(fail):
                masterLfns.pop(lfn)

        # If GUIDs are supposed to be unique check their pre-existance
        if self.db.uniqueGUID:
            fail = self._checkUniqueGUID(masterLfns, connection=connection)
            failed.update(fail)
            for lfn in fail:
                masterLfns.pop(lfn)

        # If we have files left to register
        if masterLfns:
            # Create the directories for the supplied files and store their IDs
            directories = self._getFileDirectories(list(masterLfns))
            for directory, fileNames in directories.items():
                res = self.db.dtree.makeDirectories(directory, credDict)
                if not res["OK"]:
                    for fileName in fileNames:
                        lfn = os.path.join(directory, fileName)
                        failed[lfn] = res["Message"]
                        masterLfns.pop(lfn)
                    continue
                for fileName in fileNames:
                    if not fileName:
                        failed[directory] = "Is no a valid file"
                        masterLfns.pop(directory)
                        continue

                    lfn = "%s/%s" % (directory, fileName)
                    lfn = lfn.replace("//", "/")

                    # This condition should never be true, we would not be here otherwise...
                    if not res["OK"]:
                        failed[lfn] = "Failed to create directory for file"
                        masterLfns.pop(lfn)
                    else:
                        masterLfns[lfn]["DirID"] = res["Value"]

        # If we still have files left to register
        if masterLfns:
            res = self._insertFiles(masterLfns, uid, gid, connection=connection)
            if not res["OK"]:
                for lfn in list(masterLfns):  # pylint: disable=consider-iterating-dictionary
                    failed[lfn] = res["Message"]
                    masterLfns.pop(lfn)
            else:
                for lfn, error in res["Value"]["Failed"].items():
                    failed[lfn] = error
                    masterLfns.pop(lfn)
                masterLfns = res["Value"]["Successful"]

        # Add the ancestors
        if masterLfns:
            res = self._populateFileAncestors(masterLfns, connection=connection)
            toPurge = []
            if not res["OK"]:
                for lfn in masterLfns.keys():
                    failed[lfn] = "Failed while registering ancestors"
                    toPurge.append(masterLfns[lfn]["FileID"])
            else:
                failed.update(res["Value"]["Failed"])
                for lfn, error in res["Value"]["Failed"].items():
                    toPurge.append(masterLfns[lfn]["FileID"])
            if toPurge:
                self._deleteFiles(toPurge, connection=connection)

        # Register the replicas
        newlyRegistered = {}
        if masterLfns:
            res = self._insertReplicas(masterLfns, master=True, connection=connection)
            toPurge = []
            if not res["OK"]:
                for lfn in masterLfns.keys():
                    failed[lfn] = "Failed while registering replica"
                    toPurge.append(masterLfns[lfn]["FileID"])
            else:
                newlyRegistered = res["Value"]["Successful"]
                successful.update(newlyRegistered)
                failed.update(res["Value"]["Failed"])
                for lfn, error in res["Value"]["Failed"].items():
                    toPurge.append(masterLfns[lfn]["FileID"])
            if toPurge:
                self._deleteFiles(toPurge, connection=connection)

        # Add extra replicas for successfully registered LFNs
        for lfn in list(extraLfns):
            if lfn not in successful:
                extraLfns.pop(lfn)

        if extraLfns:
            res = self._findFiles(list(extraLfns), ["FileID", "DirID"], connection=connection)
            if not res["OK"]:
                for lfn in list(lfns):
                    failed[lfn] = "Failed while registering extra replicas"
                    successful.pop(lfn)
                    extraLfns.pop(lfn)
            else:
                failed.update(res["Value"]["Failed"])
                for lfn in res["Value"]["Failed"]:
                    successful.pop(lfn)
                    extraLfns.pop(lfn)
                for lfn, fileDict in res["Value"]["Successful"].items():
                    extraLfns[lfn]["FileID"] = fileDict["FileID"]
                    extraLfns[lfn]["DirID"] = fileDict["DirID"]

            if extraLfns:
                res = self._insertReplicas(extraLfns, master=False, connection=connection)
                if not res["OK"]:
                    for lfn in extraLfns:  # pylint: disable=consider-iterating-dictionary
                        failed[lfn] = "Failed while registering extra replicas"
                        successful.pop(lfn)
                else:
                    newlyRegistered = res["Value"]["Successful"]
                    successful.update(newlyRegistered)
                    failed.update(res["Value"]["Failed"])

        return S_OK({"Successful": successful, "Failed": failed})

    def _updateDirectoryUsage(self, directorySEDict, change, connection=False):
        connection = self._getConnection(connection)
        for directoryID in directorySEDict.keys():
            result = self.db.dtree.getPathIDsByID(directoryID)
            if not result["OK"]:
                return result
            parentIDs = result["Value"]
            dirDict = directorySEDict[directoryID]
            for seID in dirDict.keys():
                seDict = dirDict[seID]
                files = seDict["Files"]
                size = seDict["Size"]
                insertTuples = []
                for dirID in parentIDs:
                    insertTuples.append("(%d,%d,%d,%d,UTC_TIMESTAMP())" % (dirID, seID, size, files))

                req = "INSERT INTO FC_DirectoryUsage (DirID,SEID,SESize,SEFiles,LastUpdate) "
                req += "VALUES %s" % ",".join(insertTuples)
                req += (
                    " ON DUPLICATE KEY UPDATE SESize=SESize%s%d, SEFiles=SEFiles%s%d, LastUpdate=UTC_TIMESTAMP() "
                    % (change, size, change, files)
                )
                res = self.db._update(req)
                if not res["OK"]:
                    gLogger.warn("Failed to update FC_DirectoryUsage", res["Message"])
        return S_OK()

    def _populateFileAncestors(self, lfns, connection=False):
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, lfnDict in lfns.items():
            originalFileID = lfnDict["FileID"]
            originalDepth = lfnDict.get("AncestorDepth", 1)
            ancestors = lfnDict.get("Ancestors", [])
            if isinstance(ancestors, six.string_types):
                ancestors = [ancestors]
            if lfn in ancestors:
                ancestors.remove(lfn)
            if not ancestors:
                successful[lfn] = True
                continue
            res = self._findFiles(ancestors, connection=connection)
            if res["Value"]["Failed"]:
                failed[lfn] = "Failed to resolve ancestor files"
                continue
            ancestorIDs = res["Value"]["Successful"]
            fileIDLFNs = {}
            toInsert = {}
            for ancestor in ancestorIDs.keys():
                fileIDLFNs[ancestorIDs[ancestor]["FileID"]] = ancestor
                toInsert[ancestorIDs[ancestor]["FileID"]] = originalDepth
            res = self._getFileAncestors(list(fileIDLFNs))
            if not res["OK"]:
                failed[lfn] = "Failed to obtain all ancestors"
                continue
            fileIDAncestorDict = res["Value"]
            for fileIDDict in fileIDAncestorDict.values():
                for ancestorID, relativeDepth in fileIDDict.items():
                    toInsert[ancestorID] = relativeDepth + originalDepth
            res = self._insertFileAncestors(originalFileID, toInsert, connection=connection)
            if not res["OK"]:
                if "Duplicate" in res["Message"]:
                    failed[lfn] = "Failed to insert ancestor files: duplicate entry"
                else:
                    failed[lfn] = "Failed to insert ancestor files"
            else:
                successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def _insertFileAncestors(self, fileID, ancestorDict, connection=False):
        connection = self._getConnection(connection)
        ancestorTuples = []
        for ancestorID, depth in ancestorDict.items():
            ancestorTuples.append("(%d,%d,%d)" % (fileID, ancestorID, depth))
        if not ancestorTuples:
            return S_OK()
        req = "INSERT INTO FC_FileAncestors (FileID, AncestorID, AncestorDepth) VALUES %s" % intListToString(
            ancestorTuples
        )
        return self.db._update(req, connection)

    def _getFileAncestors(self, fileIDs, depths=[], connection=False):
        connection = self._getConnection(connection)
        req = "SELECT FileID, AncestorID, AncestorDepth FROM FC_FileAncestors WHERE FileID IN (%s)" % intListToString(
            fileIDs
        )
        if depths:
            req = "%s AND AncestorDepth IN (%s);" % (req, intListToString(depths))
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        fileIDAncestors = {}
        for fileID, ancestorID, depth in res["Value"]:
            if fileID not in fileIDAncestors:
                fileIDAncestors[fileID] = {}
            fileIDAncestors[fileID][ancestorID] = depth
        return S_OK(fileIDAncestors)

    def _getFileDescendents(self, fileIDs, depths, connection=False):
        connection = self._getConnection(connection)
        req = (
            "SELECT AncestorID, FileID, AncestorDepth FROM FC_FileAncestors WHERE AncestorID IN (%s)"
            % intListToString(fileIDs)
        )
        if depths:
            req = "%s AND AncestorDepth IN (%s);" % (req, intListToString(depths))
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        fileIDAncestors = {}
        for ancestorID, fileID, depth in res["Value"]:
            if ancestorID not in fileIDAncestors:
                fileIDAncestors[ancestorID] = {}
            fileIDAncestors[ancestorID][fileID] = depth
        return S_OK(fileIDAncestors)

    def addFileAncestors(self, lfns, connection=False):
        """Add file ancestors to the catalog"""
        connection = self._getConnection(connection)
        failed = {}
        successful = {}
        result = self._findFiles(list(lfns), connection=connection)
        if not result["OK"]:
            return result
        if result["Value"]["Failed"]:
            failed.update(result["Value"]["Failed"])
            for lfn in result["Value"]["Failed"]:
                lfns.pop(lfn)
        if not lfns:
            return S_OK({"Successful": successful, "Failed": failed})

        for lfn in result["Value"]["Successful"]:
            lfns[lfn]["FileID"] = result["Value"]["Successful"][lfn]["FileID"]

        result = self._populateFileAncestors(lfns, connection)
        if not result["OK"]:
            return result
        failed.update(result["Value"]["Failed"])
        successful = result["Value"]["Successful"]
        return S_OK({"Successful": successful, "Failed": failed})

    def _getFileRelatives(self, lfns, depths, relation, connection=False):
        connection = self._getConnection(connection)
        failed = {}
        successful = {}
        result = self._findFiles(list(lfns), connection=connection)
        if not result["OK"]:
            return result
        if result["Value"]["Failed"]:
            failed.update(result["Value"]["Failed"])
            for lfn in result["Value"]["Failed"]:
                lfns.pop(lfn)
        if not lfns:
            return S_OK({"Successful": successful, "Failed": failed})

        inputIDDict = {}
        for lfn in result["Value"]["Successful"]:
            inputIDDict[result["Value"]["Successful"][lfn]["FileID"]] = lfn

        inputIDs = list(inputIDDict)
        if relation == "ancestor":
            result = self._getFileAncestors(inputIDs, depths, connection)
        else:
            result = self._getFileDescendents(inputIDs, depths, connection)

        if not result["OK"]:
            return result

        failed = {}
        successful = {}
        relDict = result["Value"]
        for id_ in inputIDs:
            if id_ in relDict:
                result = self._getFileLFNs(list(relDict[id_]))
                if not result["OK"]:
                    failed[inputIDDict[id]] = "Failed to find %s" % relation
                else:
                    if result["Value"]["Successful"]:
                        resDict = {}
                        for aID in result["Value"]["Successful"]:
                            resDict[result["Value"]["Successful"][aID]] = relDict[id_][aID]
                        successful[inputIDDict[id_]] = resDict
                    for aID in result["Value"]["Failed"]:
                        failed[inputIDDict[id_]] = "Failed to get the ancestor LFN"
            else:
                successful[inputIDDict[id_]] = {}

        return S_OK({"Successful": successful, "Failed": failed})

    def getFileAncestors(self, lfns, depths, connection=False):
        return self._getFileRelatives(lfns, depths, "ancestor", connection)

    def getFileDescendents(self, lfns, depths, connection=False):
        return self._getFileRelatives(lfns, depths, "descendent", connection)

    def _getExistingMetadata(self, lfns, connection=False):
        connection = self._getConnection(connection)
        # Check whether the files already exist before adding
        res = self._findFiles(lfns, ["FileID", "Size", "Checksum", "GUID"], connection=connection)
        if not res["OK"]:
            return res
        successful = res["Value"]["Successful"]
        failed = res["Value"]["Failed"]
        for lfn, error in list(failed.items()):
            if error == "No such file or directory":
                failed.pop(lfn)
        return S_OK((successful, failed))

    def _checkExistingMetadata(self, existingLfns, lfns):
        failed = {}
        successful = {}
        fileIDLFNs = {}
        for lfn, fileDict in existingLfns.items():
            fileIDLFNs[fileDict["FileID"]] = lfn
        # For those that exist get the replicas to determine whether they are already registered
        res = self._getFileReplicas(list(fileIDLFNs))
        if not res["OK"]:
            for lfn in fileIDLFNs.values():
                failed[lfn] = "Failed checking pre-existing replicas"
        else:
            replicaDict = res["Value"]
            for fileID, lfn in fileIDLFNs.items():
                fileMetadata = existingLfns[lfn]
                existingGuid = fileMetadata["GUID"]
                existingSize = fileMetadata["Size"]
                existingChecksum = fileMetadata["Checksum"]
                newGuid = lfns[lfn]["GUID"]
                newSize = lfns[lfn]["Size"]
                newChecksum = lfns[lfn]["Checksum"]
                # Ensure that the key file metadata is the same
                if (existingGuid != newGuid) or (existingSize != newSize) or (existingChecksum != newChecksum):
                    failed[lfn] = "File already registered with alternative metadata"
                # If the DB does not have replicas for this file return an error
                elif fileID not in replicaDict or not replicaDict[fileID]:
                    failed[lfn] = "File already registered with no replicas"
                # If the supplied SE is not in the existing replicas return an error
                elif not lfns[lfn]["SE"] in replicaDict[fileID].keys():
                    failed[lfn] = "File already registered with alternative replicas"
                # If we get here the file being registered already exists exactly in the DB
                else:
                    successful[lfn] = True
        return successful, failed

    def _checkUniqueGUID(self, lfns, connection=False):
        connection = self._getConnection(connection)
        guidLFNs = {}
        failed = {}
        for lfn, fileDict in lfns.items():
            guidLFNs[fileDict["GUID"]] = lfn
        res = self._getFileIDFromGUID(list(guidLFNs), connection=connection)
        if not res["OK"]:
            return dict.fromkeys(lfns, res["Message"])
        for guid, fileID in res["Value"].items():
            # resolve this to LFN
            failed[guidLFNs[guid]] = "GUID already registered for another file %s" % fileID
        return failed

    def removeFile(self, lfns, connection=False):
        connection = self._getConnection(connection)
        """ Remove file from the catalog """
        successful = {}
        failed = {}
        res = self._findFiles(lfns, ["DirID", "FileID", "Size"], connection=connection)
        if not res["OK"]:
            return res
        for lfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                successful[lfn] = True
            else:
                failed[lfn] = error
        fileIDLfns = {}
        lfns = res["Value"]["Successful"]
        for lfn, lfnDict in lfns.items():
            fileIDLfns[lfnDict["FileID"]] = lfn

        res = self._computeStorageUsageOnRemoveFile(lfns, connection=connection)
        if not res["OK"]:
            return res
        directorySESizeDict = res["Value"]

        # Now do removal
        res = self._deleteFiles(list(fileIDLfns), connection=connection)
        if not res["OK"]:
            for lfn in fileIDLfns.values():
                failed[lfn] = res["Message"]
        else:
            # Update the directory usage
            self._updateDirectoryUsage(directorySESizeDict, "-", connection=connection)
            for lfn in fileIDLfns.values():
                successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def _computeStorageUsageOnRemoveFile(self, lfns, connection=False):
        # Resolve the replicas to calculate reduction in storage usage
        fileIDLfns = {}
        for lfn, lfnDict in lfns.items():
            fileIDLfns[lfnDict["FileID"]] = lfn
        res = self._getFileReplicas(list(fileIDLfns), connection=connection)
        if not res["OK"]:
            return res
        directorySESizeDict = {}
        for fileID, seDict in res["Value"].items():
            dirID = lfns[fileIDLfns[fileID]]["DirID"]
            size = lfns[fileIDLfns[fileID]]["Size"]
            directorySESizeDict.setdefault(dirID, {})
            directorySESizeDict[dirID].setdefault(0, {"Files": 0, "Size": 0})
            directorySESizeDict[dirID][0]["Size"] += size
            directorySESizeDict[dirID][0]["Files"] += 1
            for seName in seDict.keys():
                res = self.db.seManager.findSE(seName)
                if not res["OK"]:
                    return res
                seID = res["Value"]
                size = lfns[fileIDLfns[fileID]]["Size"]
                directorySESizeDict[dirID].setdefault(seID, {"Files": 0, "Size": 0})
                directorySESizeDict[dirID][seID]["Size"] += size
                directorySESizeDict[dirID][seID]["Files"] += 1

        return S_OK(directorySESizeDict)

    def setFileStatus(self, lfns, connection=False):
        """Get set the group for the supplied files"""
        connection = self._getConnection(connection)
        res = self._findFiles(lfns, ["FileID", "UID"], connection=connection)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn in res["Value"]["Successful"]:
            status = lfns[lfn]
            if isinstance(status, six.string_types):
                if status not in self.db.validFileStatus:
                    failed[lfn] = "Invalid file status %s" % status
                    continue
                result = self._getStatusInt(status, connection=connection)
                if not result["OK"]:
                    failed[lfn] = res["Message"]
                    continue
                status = result["Value"]
            fileID = res["Value"]["Successful"][lfn]["FileID"]
            res = self._setFileParameter(fileID, "Status", status, connection=connection)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    ######################################################
    #
    # Replica write methods
    #

    def addReplica(self, lfns, connection=False):
        """Add replica to the catalog"""
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, info in list(lfns.items()):
            res = self._checkInfo(info, ["PFN", "SE"])
            if not res["OK"]:
                failed[lfn] = res["Message"]
                lfns.pop(lfn)
        res = self._addReplicas(lfns, connection=connection)
        if not res["OK"]:
            for lfn in lfns:
                failed[lfn] = res["Message"]
        else:
            failed.update(res["Value"]["Failed"])
            successful.update(res["Value"]["Successful"])
        return S_OK({"Successful": successful, "Failed": failed})

    def _addReplicas(self, lfns, connection=False):

        connection = self._getConnection(connection)
        successful = {}
        res = self._findFiles(list(lfns), ["DirID", "FileID", "Size"], connection=connection)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        for lfn in failed:
            lfns.pop(lfn)
        lfnFileIDDict = res["Value"]["Successful"]
        for lfn, fileDict in lfnFileIDDict.items():
            lfns[lfn].update(fileDict)
        res = self._insertReplicas(lfns, connection=connection)
        if not res["OK"]:
            for lfn in lfns:
                failed[lfn] = res["Message"]
        else:
            successful = res["Value"]["Successful"]
            failed.update(res["Value"]["Failed"])
        return S_OK({"Successful": successful, "Failed": failed})

    def removeReplica(self, lfns, connection=False):
        """Remove replica from catalog"""
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, info in list(lfns.items()):
            res = self._checkInfo(info, ["SE"])
            if not res["OK"]:
                failed[lfn] = res["Message"]
                lfns.pop(lfn)
        res = self._deleteReplicas(lfns, connection=connection)
        if not res["OK"]:
            for lfn in lfns.keys():
                failed[lfn] = res["Message"]
        else:
            failed.update(res["Value"]["Failed"])
            successful.update(res["Value"]["Successful"])
        return S_OK({"Successful": successful, "Failed": failed})

    def setReplicaStatus(self, lfns, connection=False):
        """Set replica status in the catalog"""
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, info in lfns.items():
            res = self._checkInfo(info, ["SE", "Status"])
            if not res["OK"]:
                failed[lfn] = res["Message"]
                continue
            status = info["Status"]
            se = info["SE"]
            res = self._findFiles([lfn], ["FileID"], connection=connection)
            if lfn not in res["Value"]["Successful"]:
                failed[lfn] = res["Value"]["Failed"][lfn]
                continue
            fileID = res["Value"]["Successful"][lfn]["FileID"]
            res = self._setReplicaStatus(fileID, se, status, connection=connection)
            if res["OK"]:
                successful[lfn] = res["Value"]
            else:
                failed[lfn] = res["Message"]
        return S_OK({"Successful": successful, "Failed": failed})

    def setReplicaHost(self, lfns, connection=False):
        """Set replica host in the catalog"""
        connection = self._getConnection(connection)
        successful = {}
        failed = {}
        for lfn, info in lfns.items():
            res = self._checkInfo(info, ["SE", "NewSE"])
            if not res["OK"]:
                failed[lfn] = res["Message"]
                continue
            newSE = info["NewSE"]
            se = info["SE"]
            res = self._findFiles([lfn], ["FileID"], connection=connection)
            if lfn not in res["Value"]["Successful"]:
                failed[lfn] = res["Value"]["Failed"][lfn]
                continue
            fileID = res["Value"]["Successful"][lfn]["FileID"]
            res = self._setReplicaHost(fileID, se, newSE, connection=connection)
            if res["OK"]:
                successful[lfn] = res["Value"]
            else:
                failed[lfn] = res["Message"]
        return S_OK({"Successful": successful, "Failed": failed})

    ######################################################
    #
    # File read methods
    #

    def exists(self, lfns, connection=False):
        """Determine whether a file exists in the catalog"""
        connection = self._getConnection(connection)
        res = self._findFiles(lfns, allStatus=True, connection=connection)
        if not res["OK"]:
            return res
        successful = res["Value"]["Successful"]
        origFailed = res["Value"]["Failed"]
        for lfn in successful:
            successful[lfn] = lfn
        failed = {}

        if self.db.uniqueGUID:
            guidList = []
            val = None
            # Try to identify if the GUID is given
            # We consider only 2 options :
            # either {lfn : guid}
            # or P lfn : {PFN : .., GUID : ..} }
            if isinstance(lfns, dict):
                val = list(lfns.values())

            # We have values, take the first to identify the type
            if val:
                val = val[0]

            if isinstance(val, dict) and "GUID" in val:
                # We are in the case {lfn : {PFN:.., GUID:..}}
                guidList = [lfns[lfn]["GUID"] for lfn in lfns]
            elif isinstance(val, six.string_types):
                # We hope that it is the GUID which is given
                guidList = list(lfns.values())

            if guidList:
                # A dict { guid: lfn to which it is supposed to be associated }
                guidToGivenLfn = dict(zip(guidList, lfns))
                res = self.getLFNForGUID(guidList, connection)
                if not res["OK"]:
                    return res
                guidLfns = res["Value"]["Successful"]
                for guid, realLfn in guidLfns.items():
                    successful[guidToGivenLfn[guid]] = realLfn

        for lfn, error in origFailed.items():
            # It could be in successful because the guid exists with another lfn
            if lfn in successful:
                continue
            if error == "No such file or directory":
                successful[lfn] = False
            else:
                failed[lfn] = error
        return S_OK({"Successful": successful, "Failed": failed})

    def isFile(self, lfns, connection=False):
        """Determine whether a path is a file in the catalog"""
        connection = self._getConnection(connection)
        # TO DO, should check whether it is a directory if it fails
        return self.exists(lfns, connection=connection)

    def getFileSize(self, lfns, connection=False):
        """Get file size from the catalog"""
        connection = self._getConnection(connection)
        # TO DO, should check whether it is a directory if it fails
        res = self._findFiles(lfns, ["Size"], connection=connection)
        if not res["OK"]:
            return res

        totalSize = 0
        for lfn in res["Value"]["Successful"]:
            size = res["Value"]["Successful"][lfn]["Size"]
            res["Value"]["Successful"][lfn] = size
            totalSize += size

        res["TotalSize"] = totalSize
        return res

    def getFileMetadata(self, lfns, connection=False):
        """Get file metadata from the catalog"""
        connection = self._getConnection(connection)
        # TO DO, should check whether it is a directory if it fails
        return self._findFiles(
            lfns,
            [
                "Size",
                "Checksum",
                "ChecksumType",
                "UID",
                "GID",
                "GUID",
                "CreationDate",
                "ModificationDate",
                "Mode",
                "Status",
            ],
            connection=connection,
        )

    def getPathPermissions(self, paths, credDict, connection=False):
        """Get the permissions for the supplied paths"""
        connection = self._getConnection(connection)
        res = self.db.ugManager.getUserAndGroupID(credDict)
        if not res["OK"]:
            return res
        uid, gid = res["Value"]
        res = self._findFiles(paths, metadata=["Mode", "UID", "GID"], connection=connection)
        if not res["OK"]:
            return res
        successful = {}
        for dirName, dirDict in res["Value"]["Successful"].items():
            mode = dirDict["Mode"]
            p_uid = dirDict["UID"]
            p_gid = dirDict["GID"]
            successful[dirName] = {}
            if p_uid == uid:
                successful[dirName]["Read"] = mode & stat.S_IRUSR
                successful[dirName]["Write"] = mode & stat.S_IWUSR
                successful[dirName]["Execute"] = mode & stat.S_IXUSR
            elif p_gid == gid:
                successful[dirName]["Read"] = mode & stat.S_IRGRP
                successful[dirName]["Write"] = mode & stat.S_IWGRP
                successful[dirName]["Execute"] = mode & stat.S_IXGRP
            else:
                successful[dirName]["Read"] = mode & stat.S_IROTH
                successful[dirName]["Write"] = mode & stat.S_IWOTH
                successful[dirName]["Execute"] = mode & stat.S_IXOTH
        return S_OK({"Successful": successful, "Failed": res["Value"]["Failed"]})

    ######################################################
    #
    # Replica read methods
    #

    def __getReplicasForIDs(self, fileIDLfnDict, allStatus, connection=False):
        """Get replicas for files with already resolved IDs"""
        replicas = {}
        if fileIDLfnDict:
            fields = []
            if not self.db.lfnPfnConvention or self.db.lfnPfnConvention == "Weak":
                fields = ["PFN"]
            res = self._getFileReplicas(
                list(fileIDLfnDict), fields_input=fields, allStatus=allStatus, connection=connection
            )
            if not res["OK"]:
                return res
            for fileID, seDict in res["Value"].items():
                lfn = fileIDLfnDict[fileID]
                replicas[lfn] = {}
                for se, repDict in seDict.items():
                    pfn = repDict.get("PFN", "")
                    replicas[lfn][se] = pfn

        result = S_OK(replicas)
        return result

    def getReplicas(self, lfns, allStatus, connection=False):
        """Get file replicas from the catalog"""
        connection = self._getConnection(connection)

        # Get FileID <-> LFN correspondence first
        res = self._findFileIDs(lfns, connection=connection)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        fileIDLFNs = {}
        for lfn, fileID in res["Value"]["Successful"].items():
            fileIDLFNs[fileID] = lfn

        result = self.__getReplicasForIDs(fileIDLFNs, allStatus, connection)
        if not result["OK"]:
            return result
        replicas = result["Value"]

        return S_OK({"Successful": replicas, "Failed": failed})

    def getReplicasByMetadata(self, metaDict, path, allStatus, credDict, connection=False):
        """Get file replicas for files corresponding to the given metadata"""
        connection = self._getConnection(connection)

        # Get FileID <-> LFN correspondence first
        failed = {}
        result = self.db.fmeta.findFilesByMetadata(metaDict, path, credDict)
        if not result["OK"]:
            return result
        idLfnDict = result["Value"]

        result = self.__getReplicasForIDs(idLfnDict, allStatus, connection)
        if not result["OK"]:
            return result
        replicas = result["Value"]

        return S_OK({"Successful": replicas, "Failed": failed})

    def getReplicaStatus(self, lfns, connection=False):
        """Get replica status from the catalog"""
        connection = self._getConnection(connection)
        res = self._findFiles(lfns, connection=connection)
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        fileIDLFNs = {}
        for lfn, fileDict in res["Value"]["Successful"].items():
            fileID = fileDict["FileID"]
            fileIDLFNs[fileID] = lfn
        successful = {}
        if fileIDLFNs:
            res = self._getFileReplicas(list(fileIDLFNs), allStatus=True, connection=connection)
            if not res["OK"]:
                return res
            for fileID, seDict in res["Value"].items():
                lfn = fileIDLFNs[fileID]
                requestedSE = lfns[lfn]
                if not requestedSE:
                    failed[lfn] = "Replica info not supplied"
                elif requestedSE not in seDict:
                    failed[lfn] = "No replica at supplied site"
                else:
                    successful[lfn] = seDict[requestedSE]["Status"]
        return S_OK({"Successful": successful, "Failed": failed})

    ######################################################
    #
    # General usage methods
    #

    def _getStatusInt(self, status, connection=False):
        connection = self._getConnection(connection)
        req = "SELECT StatusID FROM FC_Statuses WHERE Status = '%s';" % status
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        if res["Value"]:
            return S_OK(res["Value"][0][0])
        req = "INSERT INTO FC_Statuses (Status) VALUES ('%s');" % status
        res = self.db._update(req, connection)
        if not res["OK"]:
            return res
        return S_OK(res["lastRowId"])

    def _getIntStatus(self, statusID, connection=False):
        if statusID in self.statusDict:
            return S_OK(self.statusDict[statusID])
        connection = self._getConnection(connection)
        req = "SELECT StatusID,Status FROM FC_Statuses"
        res = self.db._query(req, connection)
        if not res["OK"]:
            return res
        if res["Value"]:
            for row in res["Value"]:
                self.statusDict[int(row[0])] = row[1]
        if statusID in self.statusDict:
            return S_OK(self.statusDict[statusID])
        return S_OK("Unknown")

    def getFileIDsInDirectory(self, dirID, requestString=False):
        """Get a list of IDs for all the files stored in given directories or their
            subdirectories

        :param dirID: single directory ID or a list of directory IDs
        :type dirID: int or python:list[int]
        :param bool requestString: if True return result as a SQL SELECT string
        :return: list of file IDs or SELECT string
        """
        return self._getDirectoryFileIDs(dirID, requestString=requestString)

    def getFilesInDirectory(self, dirID, verbose=False, connection=False):
        connection = self._getConnection(connection)
        files = {}
        res = self._getDirectoryFiles(
            dirID,
            [],
            [
                "FileID",
                "Size",
                "GUID",
                "Checksum",
                "ChecksumType",
                "Type",
                "UID",
                "GID",
                "CreationDate",
                "ModificationDate",
                "Mode",
                "Status",
            ],
            connection=connection,
        )
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_OK(files)
        fileIDNames = {}
        for fileName, fileDict in res["Value"].items():
            try:
                files[fileName] = {}
                files[fileName]["MetaData"] = fileDict
                fileIDNames[fileDict["FileID"]] = fileName
            except KeyError:
                # If we return S_ERROR here, it gets treated as an empty directory in most cases
                # and the user isn't actually warned
                raise Exception(
                    "File entry for '%s' is corrupt (DirID %s), please contact the catalog administrator"
                    % (fileName, dirID)
                )

        if verbose:
            result = self._getFileReplicas(list(fileIDNames), connection=connection)
            if not result["OK"]:
                return result
            for fileID, seDict in result["Value"].items():
                fileName = fileIDNames[fileID]
                files[fileName]["Replicas"] = seDict

        return S_OK(files)

    def getDirectoryReplicas(self, dirID, path, allStatus=False, connection=False):
        """Get the replicas for all the Files in the given Directory

        :param int dirID: ID of the directory
        :param unused path: useless
        :param bool allStatus: whether all replicas and file status are considered
                            If False, take the visibleFileStatus and visibleReplicaStatus values from the configuration
        """
        connection = self._getConnection(connection)
        result = self._getDirectoryReplicas(dirID, allStatus, connection)
        if not result["OK"]:
            return result

        resultDict = {}
        seDict = {}
        for fileName, fileID, seID, pfn in result["Value"]:
            resultDict.setdefault(fileName, {})
            if seID not in seDict:
                res = self.db.seManager.getSEName(seID)
                if not res["OK"]:
                    seDict[seID] = "Unknown"
                else:
                    seDict[seID] = res["Value"]
            se = seDict[seID]
            resultDict[fileName][se] = pfn

        return S_OK(resultDict)

    def _getFileDirectories(self, lfns):
        """For a list of lfn, returns a dictionary with key the directory, and value
        the files in that directory. It does not make any query, just splits the names

        :param lfns: list of lfns
        :type lfns: python:list
        """
        dirDict = {}
        for lfn in lfns:
            lfnDir = os.path.dirname(lfn)
            lfnFile = os.path.basename(lfn)
            dirDict.setdefault(lfnDir, [])
            dirDict[lfnDir].append(lfnFile)
        return dirDict

    def _checkInfo(self, info, requiredKeys):
        if not info:
            return S_ERROR("Missing parameters")
        for key in requiredKeys:
            if key not in info:
                return S_ERROR("Missing '%s' parameter" % key)
        return S_OK()

    # def _checkLFNPFNConvention( self, lfn, pfn, se ):
    #   """ Check that the PFN corresponds to the LFN-PFN convention """
    #   if pfn == lfn:
    #     return S_OK()
    #   if ( len( pfn ) < len( lfn ) ) or ( pfn[-len( lfn ):] != lfn ) :
    #     return S_ERROR( 'PFN does not correspond to the LFN convention' )
    #  return S_OK()

    def changeFileGroup(self, lfns):
        """Get set the group for the supplied files

        :param lfns: dictionary < lfn : group >
        :param int/str newGroup: optional new group/groupID the same for all the supplied lfns
        """
        res = self._findFiles(lfns, ["FileID", "GID"])
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn in res["Value"]["Successful"]:
            group = lfns[lfn]
            if isinstance(group, six.string_types):
                groupRes = self.db.ugManager.findGroup(group)
                if not groupRes["OK"]:
                    return groupRes
                group = groupRes["Value"]
            currentGroup = res["Value"]["Successful"][lfn]["GID"]
            if int(group) == int(currentGroup):
                successful[lfn] = True
            else:
                fileID = res["Value"]["Successful"][lfn]["FileID"]
                res = self._setFileParameter(fileID, "GID", group)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                else:
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def changeFileOwner(self, lfns):
        """Set the owner for the supplied files

        :param lfns: dictionary < lfn : owner >
        :param int/str newOwner: optional new user/userID the same for all the supplied lfns
        """
        res = self._findFiles(lfns, ["FileID", "UID"])
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn in res["Value"]["Successful"]:
            owner = lfns[lfn]
            if isinstance(owner, six.string_types):
                userRes = self.db.ugManager.findUser(owner)
                if not userRes["OK"]:
                    return userRes
                owner = userRes["Value"]
            currentOwner = res["Value"]["Successful"][lfn]["UID"]
            if int(owner) == int(currentOwner):
                successful[lfn] = True
            else:
                fileID = res["Value"]["Successful"][lfn]["FileID"]
                res = self._setFileParameter(fileID, "UID", owner)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                else:
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def changeFileMode(self, lfns):
        """ " Set the mode for the supplied files

        :param lfns: dictionary < lfn : mode >
        :param int newMode: optional new mode the same for all the supplied lfns
        """
        res = self._findFiles(lfns, ["FileID", "Mode"])
        if not res["OK"]:
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn in res["Value"]["Successful"]:
            mode = lfns[lfn]
            currentMode = res["Value"]["Successful"][lfn]["Mode"]
            if int(currentMode) == int(mode):
                successful[lfn] = True
            else:
                fileID = res["Value"]["Successful"][lfn]["FileID"]
                res = self._setFileParameter(fileID, "Mode", mode)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                else:
                    successful[lfn] = True
        return S_OK({"Successful": successful, "Failed": failed})

    def setFileOwner(self, path, owner):
        """Set the file owner

        :param path: file path as a string or int or list of ints or select statement
        :type path: str, int or python:list[int]
        :param owner: new user as a string or int uid
        :type owner: str or int
        """

        result = self.db.ugManager.findUser(owner)
        if not result["OK"]:
            return result

        uid = result["Value"]

        return self._setFileParameter(path, "UID", uid)

    def setFileGroup(self, path, gname):
        """Set the file group

        :param path: file path as a string or int or list of ints or select statement
        :type path: str, int or python:list[int]
        :param gname: new group as a string or int gid
        :type gname: str or int
        """

        result = self.db.ugManager.findGroup(gname)
        if not result["OK"]:
            return result

        gid = result["Value"]

        return self._setFileParameter(path, "GID", gid)

    def setFileMode(self, path, mode):
        """Set the file mode

        :param path: file path as a string or int or list of ints or select statement
        :type path: str, int or python:list[int]
        :param int mode: new mode
        """
        return self._setFileParameter(path, "Mode", mode)

    def getSEDump(self, seName):
        """
         Return all the files at a given SE, together with checksum and size

        :param seName: name of the StorageElement

        :returns: S_OK with list of tuples (lfn, checksum, size)
        """
        return S_ERROR("To be implemented on derived class")



""" Utility for loading plotting types.
    Works both for Accounting and Monitoring.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re

from DIRAC.Core.Utilities.ObjectLoader import loadObjects

from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType
from DIRAC.MonitoringSystem.Client.Types.BaseType import BaseType

__RCSID__ = "$Id$"

########################################################################


class TypeLoader(object):
    """
    .. class:: BaseType

    :param dict loaded: it stores the loaded classes
    :param str path: The location of the classes
    :param ~DIRAC.MonitoringSystem.Client.Types.BaseType.BaseType parentCls: it is the parent class
    :param regexp: regular expression...
    """

    ########################################################################
    def __init__(self, plottingFamily="Accounting"):
        """c'tor"""
        self.__loaded = {}
        if plottingFamily == "Accounting":
            self.__path = "AccountingSystem/Client/Types"
            self.__parentCls = BaseAccountingType
        elif plottingFamily == "Monitoring":
            self.__path = "MonitoringSystem/Client/Types"
            self.__parentCls = BaseType
        self.__reFilter = re.compile(r".*[a-z1-9]\.py$")

    ########################################################################
    def getTypes(self):
        """
        It returns all monitoring classes
        """
        if not self.__loaded:
            self.__loaded = loadObjects(self.__path, self.__reFilter, self.__parentCls)
        return self.__loaded



""" QualityGraph represents a Quality Map of entities as a special color schema

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
from pylab import setp
from matplotlib.colors import Normalize
import matplotlib.cm as cm
from matplotlib.colorbar import make_axes, ColorbarBase
from matplotlib.dates import date2num

from DIRAC.Core.Utilities.Graphs.PlotBase import PlotBase
from DIRAC.Core.Utilities.Graphs.GraphData import GraphData
from DIRAC.Core.Utilities.Graphs.GraphUtilities import (
    to_timestamp,
    pixelToPoint,
    PrettyDateLocator,
    PrettyDateFormatter,
    PrettyScalarFormatter,
)

__RCSID__ = "$Id$"


cdict = {
    "red": ((0.0, 1.0, 1.0), (0.5, 0.0, 0.0), (1.0, 0.0, 0.0)),
    "green": ((0.0, 0.1, 0.1), (0.5, 0.9, 0.9), (1.0, 0.7, 0.7)),
    "blue": ((0.0, 0.1, 0.1), (0.5, 0.5, 0.5), (1.0, 0.0, 0.0)),
}

# color blind
# cdict = {'red': ((0.0, .5, 0.5),
#                 (0.5, .56, 0.56),
#                 (1.0, 0.63, 0.63)),
#         'green': ((0.0, 0., 0.),
#                   (0.5, 0.5, 0.5),
#                   (1.0, 1., 1.)),
#         'blue': ((0.0, 0., 0.),
#                  (0.5, 0.315, 0.315),
#                 (1.0, 0.63, 0.63))}


class QualityMapGraph(PlotBase):

    """
    The BarGraph class is a straightforward bar graph; given a dictionary
    of values, it takes the keys as the independent variable and the values
    as the dependent variable.
    """

    def __init__(self, data, ax, prefs, *args, **kw):

        PlotBase.__init__(self, data, ax, prefs, *args, **kw)
        if isinstance(data, dict):
            self.gdata = GraphData(data)
        elif isinstance(data, type) and data.__class__ == GraphData:
            self.gdata = data
        if "span" in self.prefs:
            self.width = self.prefs["span"]
        else:
            self.width = 1.0
            if self.gdata.key_type == "time":
                nKeys = self.gdata.getNumberOfKeys()
                self.width = (max(self.gdata.all_keys) - min(self.gdata.all_keys)) / nKeys

        # redefine the look of the scale if requested
        if isinstance(self.prefs["scale_data"], dict):
            self.cbBoundaries = list()
            self.cbValues = list()

            # ColorbarBase needs sorted data
            for boundary in sorted(self.prefs["scale_data"]):
                self.cbBoundaries.append(boundary)
                self.cbValues.append(self.prefs["scale_data"][boundary])
        else:
            self.cbBoundaries = None  # set default values
            self.cbValues = None

        if isinstance(self.prefs["scale_ticks"], list):
            self.cbTicks = self.prefs["scale_ticks"]
        else:
            self.cbTicks = None  # set default value

        # Setup the colormapper to get the right colors
        self.cmap = None

        max_value = prefs.get("normalization")
        if max_value:
            self.cmap = cm.YlGnBu  # pylint: disable=no-member
        else:
            max_value = 100
            self.cmap = cm.RdYlGn  # pylint: disable=no-member

        self.norms = Normalize(0, max_value)
        mapper = cm.ScalarMappable(cmap=self.cmap, norm=self.norms)

        def get_alpha(*args, **kw):
            return 1.0

        mapper.get_alpha = get_alpha
        self.mapper = mapper

    def draw(self):

        PlotBase.draw(self)
        self.x_formatter_cb(self.ax)

        if self.gdata.isEmpty():
            return None

        # Evaluate the bar width
        width = float(self.width)
        offset = 0.0
        if self.gdata.key_type == "time":
            width = width / 86400.0

        start_plot = 0
        end_plot = 0
        if "starttime" in self.prefs and "endtime" in self.prefs:
            start_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["starttime"])))
            end_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["endtime"])))

        labels = self.gdata.getLabels()
        nKeys = self.gdata.getNumberOfKeys()
        tmp_b = []
        tmp_x = []
        tmp_y = []

        self.bars = []
        labels = self.gdata.getLabels()
        nLabel = 0
        labelNames = []
        colors = []
        xmin = None
        xmax = None
        for label, _num in labels:
            labelNames.append(label)
            for key, value, _error in self.gdata.getPlotNumData(label):

                if xmin is None or xmin > (key + offset):
                    xmin = key + offset
                if xmax is None or xmax < (key + offset):
                    xmax = key + offset

                if value is not None:
                    colors.append(self.getQualityColor(value))
                    tmp_x.append(key + offset)
                    tmp_y.append(1.0)
                    tmp_b.append(float(nLabel))

            nLabel += 1

        self.bars += self.ax.bar(tmp_x, tmp_y, bottom=tmp_b, width=width, color=colors)

        dpi = self.prefs.get("dpi", 100)
        setp(self.bars, linewidth=pixelToPoint(0.5, dpi), edgecolor="#AAAAAA")

        # pivots = keys
        # for idx in range(len(pivots)):
        #    self.coords[ pivots[idx] ] = self.bars[idx]

        ymax = float(nLabel)
        self.ax.set_xlim(xmin=0.0, xmax=xmax + width + offset)
        self.ax.set_ylim(ymin=0.0, ymax=ymax)
        if self.gdata.key_type == "time":
            if start_plot and end_plot:
                self.ax.set_xlim(xmin=start_plot, xmax=end_plot)
            else:
                self.ax.set_xlim(xmin=min(tmp_x), xmax=max(tmp_x))
        self.ax.set_yticks([i + 0.5 for i in range(nLabel)])
        self.ax.set_yticklabels(labelNames)
        setp(self.ax.get_xticklines(), markersize=0.0)
        setp(self.ax.get_yticklines(), markersize=0.0)

        cax, kw = make_axes(self.ax, orientation="vertical", fraction=0.07)
        cb = ColorbarBase(
            cax, cmap=self.cmap, norm=self.norms, boundaries=self.cbBoundaries, values=self.cbValues, ticks=self.cbTicks
        )
        cb.draw_all()
        # cb = self.ax.colorbar( self.mapper, format="%d%%",
        #  orientation='horizontal', fraction=0.04, pad=0.1, aspect=40  )
        # setp( cb.outline, linewidth=.5 )
        # setp( cb.ax.get_xticklabels(), size=10 )
        # setp( cb.ax.get_xticklabels(), family=self.prefs['font_family'] )
        # setp( cb.ax.get_xticklabels(), fontname = self.prefs['font'] )

    def getQualityColor(self, value):

        if value is None or value < 0.0:
            return "#FFFFFF"
        return self.mapper.to_rgba(value)

    def getLegendData(self):

        return None

    def x_formatter_cb(self, ax):
        if self.gdata.key_type == "string":
            smap = self.gdata.getStringMap()
            reverse_smap = {}
            for key, val in smap.items():
                reverse_smap[val] = key
            ticks = sorted(smap.values())
            ax.set_xticks([i + 0.5 for i in ticks])
            ax.set_xticklabels([reverse_smap[i] for i in ticks])
            # labels = ax.get_xticklabels()
            ax.grid(False)
            if self.log_xaxis:
                xmin = 0.001
            else:
                xmin = 0.0
            ax.set_xlim(xmin=xmin, xmax=len(ticks))
        elif self.gdata.key_type == "time":

            # ax.set_xlim( xmin=self.begin_num,xmax=self.end_num )
            dl = PrettyDateLocator()
            df = PrettyDateFormatter(dl)
            ax.xaxis.set_major_locator(dl)
            ax.xaxis.set_major_formatter(df)
            ax.xaxis.set_clip_on(False)
            sf = PrettyScalarFormatter()
            ax.yaxis.set_major_formatter(sf)
            # labels = ax.get_xticklabels()

        else:
            return None



""" StateMachine

  This module contains the basic blocks to build a state machine (State and StateMachine)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger

__RCSID__ = "$Id$"


class State(object):
    """
    State class that represents a single step on a StateMachine, with all the
    possible transitions, the default transition and an ordering level.


    examples:
      >>> s0 = State(100)
      >>> s1 = State(0, ['StateName1', 'StateName2'], defState='StateName1')
      >>> s2 = State(0, ['StateName1', 'StateName2'])
      # this example is tricky. The transition rule says that will go to
      # nextState, e.g. 'StateNext'. But, it is not on the stateMap, and there
      # is no default defined, so it will end up going to StateNext anyway. You
      # must be careful while defining states and their stateMaps and defaults.
    """

    def __init__(self, level, stateMap=None, defState=None):
        """
        :param int level: each state is mapped to an integer, which is used to sort the states according to that integer.
        :param list stateMap: it is a list (of strings) with the reachable states from this particular status.
                              If not defined, we assume there are no restrictions.
        :param str defState: default state used in case the next state is not in stateMap (not defined or simply not there).
        """

        self.level = level
        self.stateMap = stateMap if stateMap else []
        self.default = defState

    def transitionRule(self, nextState):
        """
        Method that selects next state, knowing the default and the transitions
        map, and the proposed next state. If <nextState> is in stateMap, goes there.
        If not, then goes to <self.default> if any. Otherwise, goes to <nextState>
        anyway.

        examples:
          >>> s0.transitionRule('nextState')
              'nextState'
          >>> s1.transitionRule('StateName2')
              'StateName2'
          >>> s1.transitionRule('StateNameNotInMap')
              'StateName1'
          >>> s2.transitionRule('StateNameNotInMap')
              'StateNameNotInMap'

        :param str nextState: name of the state in the stateMap
        :return: state name
        :rtype: str
        """

        # If next state is on the list of next states, go ahead.
        if nextState in self.stateMap:
            return nextState

        # If not, calculate defaultState:
        # if there is a default, that one
        # otherwise is nextState (states with empty list have no movement restrictions)
        defaultNext = self.default if self.default else nextState
        return defaultNext


class StateMachine(object):
    """
    StateMachine class that represents the whole state machine with all transitions.

    examples:
      >>> sm0 = StateMachine()
      >>> sm1 = StateMachine(state = 'Active')

    :param state: current state of the StateMachine, could be None if we do not use the
        StateMachine to calculate transitions. Beware, it is not checked if the
        state is on the states map !
    :type state: None or str

    """

    def __init__(self, state=None):
        """
        Constructor.
        """

        self.state = state
        # To be overwritten by child classes, unless you like Nirvana state that much.
        self.states = {"Nirvana": State(100)}

    def getLevelOfState(self, state):
        """
        Given a state name, it returns its level (integer), which defines the hierarchy.

        >>> sm0.getLevelOfState('Nirvana')
            100
        >>> sm0.getLevelOfState('AnotherState')
            -1

        :param str state: name of the state, it should be on <self.states> key set
        :return: `int` || -1 (if not in <self.states>)
        """

        if state not in self.states:
            return -1
        return self.states[state].level

    def setState(self, candidateState, noWarn=False):
        """Makes sure the state is either None or known to the machine, and that it is a valid state to move into.
            Final states are also checked.

        examples:
          >>> sm0.setState(None)['OK']
              True
          >>> sm0.setState('Nirvana')['OK']
              True
          >>> sm0.setState('AnotherState')['OK']
              False

        :param state: state which will be set as current state of the StateMachine
        :type state: None or str
        :return: S_OK || S_ERROR
        """
        if candidateState == self.state:
            return S_OK(candidateState)

        if not candidateState:
            self.state = candidateState
        elif candidateState in self.states:
            if not self.states[self.state].stateMap:
                if not noWarn:
                    gLogger.warn("Final state, won't move", "(%s, asked to move to %s)" % (self.state, candidateState))
                return S_OK(self.state)
            if candidateState not in self.states[self.state].stateMap:
                gLogger.warn("Can't move from %s to %s, choosing a good one" % (self.state, candidateState))
            result = self.getNextState(candidateState)
            if not result["OK"]:
                return result
            self.state = result["Value"]
            # If the StateMachine does not accept the candidate, return error message
        else:
            return S_ERROR("setState: %r is not a valid state" % candidateState)

        return S_OK(self.state)

    def getStates(self):
        """
        Returns all possible states in the state map

        examples:
          >>> sm0.getStates()
              [ 'Nirvana' ]

        :return: list(stateNames)
        """

        return list(self.states)

    def getNextState(self, candidateState):
        """
        Method that gets the next state, given the proposed transition to candidateState.
        If candidateState is not on the state map <self.states>, it is rejected. If it is
        not the case, we have two options: if <self.state> is None, then the next state
        will be <candidateState>. Otherwise, the current state is using its own
        transition rule to decide.

        examples:
          >>> sm0.getNextState(None)
              S_OK(None)
          >>> sm0.getNextState('NextState')
              S_OK('NextState')

        :param str candidateState: name of the next state
        :return: S_OK(nextState) || S_ERROR
        """
        if candidateState not in self.states:
            return S_ERROR("getNextState: %r is not a valid state" % candidateState)

        # FIXME: do we need this anymore ?
        if self.state is None:
            return S_OK(candidateState)

        return S_OK(self.states[self.state].transitionRule(candidateState))



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-pilot-info
# Author :  Ricardo Graciani
########################################################################
"""
Retrieve available info about the given pilot

Example:
  $ dirac-admin-get-pilot-info https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw
  {'https://marlb.in2p3.fr:9000/26KCLKBFtxXKHF4_ZrQjkw': {'AccountingSent': 'False',
                                                          'BenchMark': 0.0,
                                                          'Broker': 'marwms.in2p3.fr',
                                                          'DestinationSite': 'cclcgceli01.in2p3.fr',
                                                          'GridSite': 'LCG.IN2P3.fr',
                                                          'GridType': 'gLite',
                                                          'LastUpdateTime': datetime.datetime(2011, 2, 21, 12, 49, 14),
                                                          'OutputReady': 'False',
                                                          'OwnerDN': '/O=GRID/C=FR/O=CNRS/OU=LPC/CN=Sebastien Guizard',
                                                          'OwnerGroup': '/biomed',
                                                          'ParentID': 0L,
                                                          'PilotID': 2241L,
                                                          'PilotJobReference': 'https://marlb.in2p3.fr:9000/2KHFrQjkw',
                                                          'PilotStamp': '',
                                                          'Status': 'Done',
                                                          'SubmissionTime': datetime.datetime(2011, 2, 21, 12, 27, 52),
                                                          'TaskQueueID': 399L}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# pylint: disable=wrong-import-position
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

extendedPrint = False


def setExtendedPrint(_arg):
    global extendedPrint
    extendedPrint = True


@Script()
def main():
    Script.registerSwitch("e", "extended", "Get extended printout", setExtendedPrint)
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    dirac = Dirac()
    exitCode = 0
    errorList = []

    for gridID in args:
        result = diracAdmin.getPilotInfo(gridID)
        if not result["OK"]:
            errorList.append((gridID, result["Message"]))
            exitCode = 2
        else:
            res = result["Value"][gridID]
            if extendedPrint:
                tab = ""
                for key in [
                    "PilotJobReference",
                    "Status",
                    "OwnerDN",
                    "OwnerGroup",
                    "SubmissionTime",
                    "DestinationSite",
                    "GridSite",
                ]:
                    if key in res:
                        diracAdmin.log.notice("%s%s: %s" % (tab, key, res[key]))
                        if not tab:
                            tab = "  "
                diracAdmin.log.notice("")
                for jobID in res["Jobs"]:
                    tab = "  "
                    result = dirac.getJobAttributes(int(jobID))
                    if not result["OK"]:
                        errorList.append((gridID, result["Message"]))
                        exitCode = 2
                    else:
                        job = result["Value"]
                        diracAdmin.log.notice("%sJob ID: %s" % (tab, jobID))
                        tab += "  "
                        for key in [
                            "OwnerDN",
                            "OwnerGroup",
                            "JobName",
                            "Status",
                            "StartExecTime",
                            "LastUpdateTime",
                            "EndExecTime",
                        ]:
                            if key in job:
                                diracAdmin.log.notice("%s%s:" % (tab, key), job[key])
                diracAdmin.log.notice("")
            else:
                print(diracAdmin.pPrint.pformat({gridID: res}))

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



""" DIPStorage class is the client of the DIRAC Storage Element.

    The following methods are available in the Service interface

    getMetadata()
    get()
    getDir()
    put()
    putDir()
    remove()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import random

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.Utilities.File import getSize


class DIPStorage(StorageBase):

    _INPUT_PROTOCOLS = ["file", "dip", "dips"]
    _OUTPUT_PROTOCOLS = ["dip", "dips"]

    def __init__(self, storageName, parameters):
        """ """
        StorageBase.__init__(self, storageName, parameters)
        self.pluginName = "DIP"

        self.log = gLogger.getSubLogger("DIPStorage")

        # Several ports can be specified as comma separated list, choose
        # randomly one of those ports
        ports = self.protocolParameters["Port"].split(",")
        random.shuffle(ports)
        self.protocolParameters["Port"] = ports[0]

        pathDict = dict(self.protocolParameters)
        pathDict["Path"] = self.basePath
        result = pfnunparse(pathDict)
        if result["OK"]:
            self.url = result["Value"]

        self.checkSum = "CheckSum"
        self.isok = True

    def setParameters(self, parameters):
        """Applying extra storage parameters"""
        StorageBase.setParameters(self, parameters)
        if "CheckSum" in parameters and parameters["CheckSum"].lower() in ["0", "no", "false", "off"]:
            self.checkSum = "NoCheckSum"
        return S_OK()

    #############################################################
    #
    # These are the methods for file manipulation
    #

    def exists(self, path):
        """Check if the given path exists. The 'path' variable can be a string or a list of strings."""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        serviceClient = RPCClient(self.url)
        for url in urls:
            gLogger.debug("DIPStorage.exists: Determining existence of %s." % url)
            res = serviceClient.exists(url)
            if res["OK"]:
                successful[url] = res["Value"]
            else:
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def putFile(self, path, sourceSize=0):
        """Put a file to the physical storage"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        for dest_url, src_file in urls.items():
            gLogger.debug("DIPStorage.putFile: Executing transfer of %s to %s" % (src_file, dest_url))
            res = self.__putFile(src_file, dest_url)
            if res["OK"]:
                successful[dest_url] = res["Value"]
            else:
                failed[dest_url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __putFile(self, src_file, dest_url):
        res = pfnparse(src_file)
        if not res["OK"]:
            return res
        localCache = False
        srcDict = res["Value"]
        if srcDict["Protocol"] in ["dips", "dip"]:
            # Make the service URL from the file URL by stripping off the file part
            serviceDict = dict(srcDict)
            serviceDict["Path"] = "/".join(srcDict["Path"].split("/")[:3])
            serviceDict["FileName"] = ""
            res = pfnunparse(serviceDict)
            if not res["OK"]:
                return res
            srcSEURL = res["Value"]
            localCache = True
            transferClient = TransferClient(srcSEURL)
            res = transferClient.receiveFile(srcDict["FileName"], os.path.join(srcDict["Path"], srcDict["FileName"]))
            if not res["OK"]:
                return res
            src_file = srcDict["FileName"]

        if not os.path.exists(src_file):
            errStr = "DIPStorage.__putFile: The source local file does not exist."
            gLogger.error(errStr, src_file)
            return S_ERROR(errStr)
        sourceSize = getSize(src_file)
        if sourceSize == -1:
            errStr = "DIPStorage.__putFile: Failed to get file size."
            gLogger.error(errStr, src_file)
            return S_ERROR(errStr)
        transferClient = TransferClient(self.url)
        res = transferClient.sendFile(src_file, dest_url, token=self.checkSum)
        if localCache:
            os.unlink(src_file)
        if res["OK"]:
            return S_OK(sourceSize)
        return res

    def getFile(self, path, localPath=False):
        """Get a local copy in the current directory of a physical file specified by its path"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        for src_url in urls:
            fileName = os.path.basename(src_url)
            if localPath:
                dest_file = "%s/%s" % (localPath, fileName)
            else:
                dest_file = "%s/%s" % (os.getcwd(), fileName)
            gLogger.debug("DIPStorage.getFile: Executing transfer of %s to %s" % (src_url, dest_file))
            res = self.__getFile(src_url, dest_file)
            if res["OK"]:
                successful[src_url] = res["Value"]
            else:
                failed[src_url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getFile(self, src_url, dest_file):
        transferClient = TransferClient(self.url)
        res = transferClient.receiveFile(dest_file, src_url, token=self.checkSum)
        if not res["OK"]:
            return res
        if not os.path.exists(dest_file):
            errStr = "DIPStorage.__getFile: The destination local file does not exist."
            gLogger.error(errStr, dest_file)
            return S_ERROR(errStr)
        destSize = getSize(dest_file)
        if destSize == -1:
            errStr = "DIPStorage.__getFile: Failed to get the local file size."
            gLogger.error(errStr, dest_file)
            return S_ERROR(errStr)
        return S_OK(destSize)

    def removeFile(self, path):
        """Remove physically the file specified by its path"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        if not len(urls) > 0:
            return S_ERROR("DIPStorage.removeFile: No surls supplied.")
        successful = {}
        failed = {}
        serviceClient = RPCClient(self.url)
        for url in urls:
            gLogger.debug("DIPStorage.removeFile: Attempting to remove %s." % url)
            res = serviceClient.remove(url, "")
            if res["OK"]:
                successful[url] = True
            else:
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def isFile(self, path):
        """Determine whether the path is a directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.isFile: Attempting to determine whether %s paths are files." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.getMetadata(url)
            if res["OK"]:
                if res["Value"]["Exists"]:
                    if res["Value"]["Type"] == "File":
                        gLogger.debug("DIPStorage.isFile: Successfully obtained metadata for %s." % url)
                        successful[url] = True
                    else:
                        successful[url] = False
                else:
                    failed[url] = "File does not exist"
            else:
                gLogger.error("DIPStorage.isFile: Failed to get metadata for url", "%s: %s" % (url, res["Message"]))
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getFileSize(self, path):
        """Get size of supplied files"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.getFileSize: Attempting to obtain size for %s files." % len(urls))
        res = self.getFileMetadata(urls)
        if not res["OK"]:
            return res
        for url, urlDict in res["Value"]["Successful"].items():
            if urlDict["Exists"]:
                successful[url] = urlDict["Size"]
            else:
                failed[url] = "File does not exist"
        for url, error in res["Value"]["Failed"].items():
            failed[url] = error
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getFileMetadata(self, path):
        """Get metadata associated to the file"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.getFileMetadata: Attempting to obtain metadata for %s files." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            pfn = url
            if url.find(self.url) == 0:
                pfn = url[(len(self.url)) :]
            res = serviceClient.getMetadata(pfn)
            if res["OK"]:
                if res["Value"]["Exists"]:
                    if res["Value"]["Type"] == "File":
                        gLogger.debug("DIPStorage.getFileMetadata: Successfully obtained metadata for %s." % url)
                        successful[url] = res["Value"]
                    else:
                        failed[url] = "Supplied path is not a file"
                else:
                    failed[url] = "File does not exist"
            else:
                gLogger.error(
                    "DIPStorage.getFileMetadata: Failed to get metadata for url", "%s: %s" % (url, res["Message"])
                )
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    #############################################################
    #
    # These are the methods for directory manipulation
    #

    def listDirectory(self, path):
        """List the contents of the directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.listDirectory: Attempting to list %s directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.listDirectory(url, "l")
            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                files = {}
                subDirs = {}
                for subPath, pathDict in res["Value"].items():
                    if pathDict["Type"] == "File":
                        files[subPath] = pathDict
                    elif pathDict["Type"] == "Directory":
                        subDirs[subPath] = pathDict
                successful[url] = {}
                successful[url]["SubDirs"] = subDirs
                successful[url]["Files"] = files
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def isDirectory(self, path):
        """Determine whether the path is a directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.isDirectory: Attempting to determine whether %s paths are directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.getMetadata(url)
            if res["OK"]:
                if res["Value"]["Exists"]:
                    if res["Value"]["Type"] == "Directory":
                        gLogger.debug("DIPStorage.isDirectory: Successfully obtained metadata for %s." % url)
                        successful[url] = True
                    else:
                        successful[url] = False
                else:
                    failed[url] = "Path does not exist"
            else:
                gLogger.error(
                    "DIPStorage.isDirectory: Failed to get metadata for url", "%s: %s" % (url, res["Message"])
                )
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectorySize(self, path):
        """Get the size of the contents of the directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.isDirectory: Attempting to determine whether %s paths are directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.getDirectorySize(url)
            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = {"Files": 0, "Size": res["Value"], "SubDirs": 0}
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectoryMetadata(self, path):
        """Get metadata associated to the directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.getFileMetadata: Attempting to obtain metadata for %s directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.getMetadata(url)
            if res["OK"]:
                if res["Value"]["Exists"]:
                    if res["Value"]["Type"] == "Directory":
                        res["Value"]["Directory"] = True
                        gLogger.debug("DIPStorage.getFileMetadata: Successfully obtained metadata for %s." % url)
                        successful[url] = res["Value"]
                    else:
                        failed[url] = "Supplied path is not a directory"
                else:
                    failed[url] = "Directory does not exist"
            else:
                gLogger.error(
                    "DIPStorage.getFileMetadata: Failed to get metadata for url", "%s: %s" % (url, res["Message"])
                )
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def createDirectory(self, path):
        """Create the remote directory"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.createDirectory: Attempting to create %s directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.createDirectory(url)
            if res["OK"]:
                gLogger.debug("DIPStorage.createDirectory: Successfully created directory on storage: %s" % url)
                successful[url] = True
            else:
                gLogger.error(
                    "DIPStorage.createDirectory: Failed to create directory on storage.",
                    "%s: %s" % (url, res["Message"]),
                )
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def putDirectory(self, path):
        """Put a local directory to the physical storage together with all its files and subdirectories."""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.putDirectory: Attemping to put %s directories to remote storage." % len(urls))
        transferClient = TransferClient(self.url)
        for destDir, sourceDir in urls.items():
            tmpList = os.listdir(sourceDir)
            sourceFiles = ["%s/%s" % (sourceDir, x) for x in tmpList]
            res = transferClient.sendBulk(sourceFiles, destDir)
            if res["OK"]:
                successful[destDir] = {"Files": 0, "Size": 0}
            else:
                failed[destDir] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def removeDirectory(self, path, recursive=False):
        """Remove a directory from the storage together with all its files and subdirectories."""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}
        gLogger.debug("DIPStorage.removeDirectory: Attempting to remove %s directories." % len(urls))
        serviceClient = RPCClient(self.url)
        for url in urls:
            res = serviceClient.removeDirectory(url, "")
            if res["OK"]:
                gLogger.debug("DIPStorage.removeDirectory: Successfully removed directory on storage: %s" % url)
                successful[url] = {"FilesRemoved": 0, "SizeRemoved": 0}
            else:
                gLogger.error(
                    "DIPStorage.removeDirectory: Failed to remove directory from storage.",
                    "%s: %s" % (url, res["Message"]),
                )
                failed[url] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getDirectory(self, path, localPath=False):
        """Get a local copy in the current directory of a physical file specified by its path"""
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}
        gLogger.debug("DIPStorage.getDirectory: Attempting to get local copies of %s directories." % len(urls))
        transferClient = TransferClient(self.url)
        for src_dir in urls:
            if localPath:
                dest_dir = localPath
            else:
                dest_dir = os.getcwd()
            if not os.path.exists(dest_dir):
                os.mkdir(dest_dir)
            res = transferClient.receiveBulk(dest_dir, src_dir)
            if res["OK"]:
                gLogger.debug("DIPStorage.getDirectory: Successfully got local copy of %s" % src_dir)
                successful[src_dir] = {"Files": 0, "Size": 0}
            else:
                gLogger.error("DIPStorage.getDirectory: Failed to get entire directory.", src_dir)
                failed[src_dir] = res["Message"]
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def getOccupancy(self, *parms, **kws):
        """Gets the DIPStorage occupancy info

        :return: S_OK/S_ERROR (free and total space, in MB)
        """

        rpc = RPCClient(self.url, timeout=120)

        free = rpc.getFreeDiskSpace()
        if not free["OK"]:
            return free

        total = rpc.getTotalDiskSpace()
        if not total["OK"]:
            return total

        return S_OK({"Free": free["Value"], "Total": total["Value"]})



"""
:mod:  ReqClient

.. module:  ReqClient
  :synopsis: implementation of client for RequestDB using DISET framework

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import time
import random
import json
import datetime

# # from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.List import randomize, fromChar
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient


@createClient("RequestManagement/ReqManager")
class ReqClient(Client):
    """ReqClient is a class manipulating and operation on Requests."""

    __requestProxiesDict = {}
    __requestValidator = None

    def __init__(self, url=None, **kwargs):
        """c'tor

        :param self: self reference
        :param url: url of the ReqManager
        :param kwargs: forwarded to the Base Client class
        """

        super(ReqClient, self).__init__(**kwargs)
        self.serverURL = "RequestManagement/ReqManager" if not url else url

        self.log = gLogger.getSubLogger("RequestManagement/ReqClient/pid_%s" % (os.getpid()))

    def requestProxies(self, timeout=120):
        """get request proxies dict"""
        # Forward all the connection options to the requestClient
        # (e.g. the userDN to use)
        kwargs = self.getClientKWArgs()
        kwargs["timeout"] = timeout

        if not self.__requestProxiesDict:
            self.__requestProxiesDict = {}
            proxiesURLs = fromChar(PathFinder.getServiceURL("RequestManagement/ReqProxyURLs"))
            if not proxiesURLs:
                self.log.warn("CS option RequestManagement/ReqProxyURLs is not set!")
            for proxyURL in proxiesURLs:
                self.log.debug("creating RequestProxy for url = %s" % proxyURL)
                pc = Client(**kwargs)
                pc.setServer(proxyURL)
                self.__requestProxiesDict[proxyURL] = pc

        return self.__requestProxiesDict

    def requestValidator(self):
        """get request validator"""
        if not self.__requestValidator:
            self.__requestValidator = RequestValidator()
        return self.__requestValidator

    def putRequest(self, request, useFailoverProxy=True, retryMainService=0):
        """Put request to RequestManager

        :param self: self reference
        :param ~Request.Request request: Request instance
        :param bool useFailoverProxy: if False, will not attempt to forward the request to ReqProxies
        :param int retryMainService: Amount of time we retry on the main ReqHandler in case of failures

        :return: S_OK/S_ERROR
        """
        errorsDict = {"OK": False}
        valid = self.requestValidator().validate(request)
        if not valid["OK"]:
            self.log.error("putRequest: request not valid", "%s" % valid["Message"])
            return valid
        # # dump to json
        requestJSON = request.toJSON()
        if not requestJSON["OK"]:
            return requestJSON
        requestJSON = requestJSON["Value"]

        retryMainService += 1

        while retryMainService:
            retryMainService -= 1
            setRequestMgr = self._getRPC().putRequest(requestJSON)
            if setRequestMgr["OK"]:
                return setRequestMgr
            errorsDict["RequestManager"] = setRequestMgr["Message"]
            # sleep a bit
            time.sleep(random.randint(1, 5))

        self.log.warn(
            "putRequest: unable to set request '%s' at RequestManager" % request.RequestName, setRequestMgr["Message"]
        )
        proxies = self.requestProxies() if useFailoverProxy else {}
        for proxyURL in randomize(proxies.keys()):
            proxyClient = proxies[proxyURL]
            self.log.debug("putRequest: trying RequestProxy at %s" % proxyURL)
            setRequestProxy = proxyClient.putRequest(requestJSON)
            if setRequestProxy["OK"]:
                if setRequestProxy["Value"]["set"]:
                    self.log.info(
                        "putRequest: request '%s' successfully set using RequestProxy %s"
                        % (request.RequestName, proxyURL)
                    )
                elif setRequestProxy["Value"]["saved"]:
                    self.log.info(
                        "putRequest: request '%s' successfully forwarded to RequestProxy %s"
                        % (request.RequestName, proxyURL)
                    )
                return setRequestProxy
            else:
                self.log.warn(
                    "putRequest: unable to set request using RequestProxy %s: %s"
                    % (proxyURL, setRequestProxy["Message"])
                )
                errorsDict["RequestProxy(%s)" % proxyURL] = setRequestProxy["Message"]
        # # if we're here neither requestManager nor requestProxy were successful
        self.log.error("putRequest: unable to set request", "'%s'" % request.RequestName)
        errorsDict["Message"] = "ReqClient.putRequest: unable to set request '%s'" % request.RequestName
        return errorsDict

    def getRequest(self, requestID=0):
        """Get request from RequestDB

        :param self: self reference
        :param int requestID: ID of the request. If 0, choice is made for you

        :return: S_OK( Request instance ) or S_OK() or S_ERROR
        """
        self.log.debug("getRequest: attempting to get request.")
        getRequest = self._getRPC().getRequest(requestID)
        if not getRequest["OK"]:
            self.log.error("getRequest: unable to get request", "'%s' %s" % (requestID, getRequest["Message"]))
            return getRequest
        if not getRequest["Value"]:
            return getRequest
        return S_OK(Request(getRequest["Value"]))

    @ignoreEncodeWarning
    def getBulkRequests(self, numberOfRequest=10, assigned=True):
        """get bulk requests from RequestDB

        :param self: self reference
        :param str numberOfRequest: size of the bulk (default 10)

        :return: S_OK( Successful : { requestID, RequestInstance }, Failed : message  ) or S_ERROR
        """
        self.log.debug("getRequests: attempting to get request.")
        getRequests = self._getRPC().getBulkRequests(numberOfRequest, assigned)
        if not getRequests["OK"]:
            self.log.error("getRequests: unable to get '%s' requests: %s" % (numberOfRequest, getRequests["Message"]))
            return getRequests
        # No Request returned
        if not getRequests["Value"]:
            return getRequests
        # No successful Request
        if not getRequests["Value"]["Successful"]:
            return getRequests

        jsonReq = getRequests["Value"]["Successful"]
        # Do not forget to cast back str keys to int
        reqInstances = {int(rId): Request(jsonReq[rId]) for rId in jsonReq}
        failed = strToIntDict(getRequests["Value"]["Failed"])
        return S_OK({"Successful": reqInstances, "Failed": failed})

    def peekRequest(self, requestID):
        """peek request"""
        self.log.debug("peekRequest: attempting to get request.")
        peekRequest = self._getRPC().peekRequest(int(requestID))
        if not peekRequest["OK"]:
            self.log.error(
                "peekRequest: unable to peek request", "request: '%s' %s" % (requestID, peekRequest["Message"])
            )
            return peekRequest
        if not peekRequest["Value"]:
            return peekRequest
        return S_OK(Request(peekRequest["Value"]))

    def deleteRequest(self, requestID):
        """delete request given it's ID

        :param self: self reference
        :param str requestID: request ID
        """
        requestID = int(requestID)
        self.log.debug("deleteRequest: attempt to delete '%s' request" % requestID)
        deleteRequest = self._getRPC().deleteRequest(requestID)
        if not deleteRequest["OK"]:
            self.log.error(
                "deleteRequest: unable to delete request", "'%s' request: %s" % (requestID, deleteRequest["Message"])
            )
        return deleteRequest

    def getRequestIDsList(self, statusList=None, limit=None, since=None, until=None, getJobID=False):
        """get at most :limit: request ids with statuses in :statusList:"""
        statusList = statusList if statusList else list(Request.FINAL_STATES)
        limit = limit if limit else 100
        since = since.strftime("%Y-%m-%d") if since else ""
        until = until.strftime("%Y-%m-%d") if until else ""

        return self._getRPC().getRequestIDsList(statusList, limit, since, until, getJobID)

    def getScheduledRequest(self, operationID):
        """get scheduled request given its scheduled OperationID"""
        self.log.debug("getScheduledRequest: attempt to get scheduled request...")
        scheduled = self._getRPC().getScheduledRequest(operationID)
        if not scheduled["OK"]:
            self.log.error("getScheduledRequest failed", scheduled["Message"])
            return scheduled
        if scheduled["Value"]:
            return S_OK(Request(scheduled["Value"]))
        return scheduled

    def getDBSummary(self):
        """Get the summary of requests in the RequestDBs."""
        self.log.debug("getDBSummary: attempting to get RequestDB summary.")
        dbSummary = self._getRPC().getDBSummary()
        if not dbSummary["OK"]:
            self.log.error("getDBSummary: unable to get RequestDB summary", dbSummary["Message"])
        return dbSummary

    def getDigest(self, requestID):
        """Get the request digest given a request ID.

        :param self: self reference
        :param str requestID: request id
        """
        self.log.debug("getDigest: attempting to get digest for '%s' request." % requestID)
        digest = self._getRPC().getDigest(int(requestID))
        if not digest["OK"]:
            self.log.error(
                "getDigest: unable to get digest for request", "request: '%s' %s" % (requestID, digest["Message"])
            )

        return digest

    def getRequestStatus(self, requestID):
        """Get the request status given a request id.

        :param self: self reference
        :param int requestID: id of the request
        """
        if isinstance(requestID, six.string_types):
            requestID = int(requestID)
        self.log.debug("getRequestStatus: attempting to get status for '%d' request." % requestID)
        requestStatus = self._getRPC().getRequestStatus(requestID)
        if not requestStatus["OK"]:
            self.log.error(
                "getRequestStatus: unable to get status for request",
                ": '%d' %s" % (requestID, requestStatus["Message"]),
            )
        return requestStatus

    #   def getRequestName( self, requestID ):
    #     """ get request name for a given requestID """
    #     return self._getRPC().getRequestName( requestID )

    def getRequestInfo(self, requestID):
        """The the request info given a request id.

        :param self: self reference
        :param int requestID: request nid
        """
        self.log.debug("getRequestInfo: attempting to get info for '%s' request." % requestID)
        requestInfo = self._getRPC().getRequestInfo(int(requestID))
        if not requestInfo["OK"]:
            self.log.error(
                "getRequestInfo: unable to get status for request",
                "request: '%s' %s" % (requestID, requestInfo["Message"]),
            )
        return requestInfo

    def getRequestFileStatus(self, requestID, lfns):
        """Get file status for request given a request id.

        :param self: self reference
        :param int requestID: request id
        :param lfns: list of LFNs
        :type lfns: python:list
        """
        self.log.debug("getRequestFileStatus: attempting to get file statuses for '%s' request." % requestID)
        fileStatus = self._getRPC().getRequestFileStatus(int(requestID), lfns)
        if not fileStatus["OK"]:
            self.log.verbose(
                "getRequestFileStatus: unable to get file status for request",
                "request: '%s' %s" % (requestID, fileStatus["Message"]),
            )
        return fileStatus

    def finalizeRequest(self, requestID, jobID, useCertificates=True):
        """check request status and perform finalization if necessary
            update the request status and the corresponding job parameter

        :param self: self reference
        :param str requestID: request id
        :param int jobID: job id
        """

        stateServer = JobStateUpdateClient(useCertificates=useCertificates)

        # Checking if to update the job status - we should fail here, so it will be re-tried later
        # Checking the state, first
        res = self.getRequestStatus(requestID)
        if not res["OK"]:
            self.log.error(
                "finalizeRequest: failed to get request", "request: %s status: %s" % (requestID, res["Message"])
            )
            return res
        if res["Value"] != "Done":
            return S_ERROR(
                "The request %s isn't 'Done' but '%s', this should never happen, why are we here?"
                % (requestID, res["Value"])
            )

        # The request is 'Done', let's update the job status. If we fail, we should re-try later

        monitorServer = JobMonitoringClient(useCertificates=useCertificates)
        res = monitorServer.getJobSummary(int(jobID))
        if not res["OK"]:
            self.log.error("finalizeRequest: Failed to get job status", "JobID: %d" % jobID)
            return res
        elif not res["Value"]:
            self.log.info("finalizeRequest: job %d does not exist (anymore): finalizing" % jobID)
            return S_OK()
        else:
            jobStatus = res["Value"]["Status"]
            jobMinorStatus = res["Value"]["MinorStatus"]
            jobAppStatus = ""
            newJobStatus = ""
            if jobStatus == JobStatus.STALLED:
                # If job is stalled, find the previous status from the logging info
                res = monitorServer.getJobLoggingInfo(int(jobID))
                if not res["OK"]:
                    self.log.error("finalizeRequest: Failed to get job logging info", "JobID: %d" % jobID)
                    return res
                # Check the last status was Stalled and get the one before
                if len(res["Value"]) >= 2 and res["Value"][-1][0] == JobStatus.STALLED:
                    jobStatus, jobMinorStatus, jobAppStatus = res["Value"][-2][:3]
                    newJobStatus = jobStatus

            # update the job pending request digest in any case since it is modified
            self.log.info("finalizeRequest: Updating request digest for job %d" % jobID)

            digest = self.getDigest(requestID)
            if digest["OK"]:
                digest = digest["Value"]
                self.log.verbose(digest)
                res = stateServer.setJobParameter(jobID, "PendingRequest", digest)
                if not res["OK"]:
                    self.log.info("finalizeRequest: Failed to set job %d parameter: %s" % (jobID, res["Message"]))
                    return res
            else:
                self.log.error(
                    "finalizeRequest: Failed to get request digest for %s: %s" % (requestID, digest["Message"])
                )
            if jobStatus == JobStatus.COMPLETED:
                # What to do? Depends on what we have in the minorStatus
                if jobMinorStatus == JobMinorStatus.PENDING_REQUESTS:
                    newJobStatus = JobStatus.DONE
                elif jobMinorStatus == JobMinorStatus.APP_ERRORS:
                    newJobStatus = JobStatus.FAILED
                elif jobMinorStatus == JobMinorStatus.MARKED_FOR_TERMINATION:
                    # If the job has been Killed, set it Killed
                    newJobStatus = JobStatus.KILLED
                else:
                    self.log.error(
                        "finalizeRequest: Unexpected jobMinorStatus", "for %d (got %s)" % (jobID, jobMinorStatus)
                    )
                    return S_ERROR("Unexpected jobMinorStatus")

            if newJobStatus:
                self.log.info("finalizeRequest: Updating job status for %d to %s/Requests done" % (jobID, newJobStatus))
            else:
                self.log.info(
                    "finalizeRequest: Updating job minor status",
                    "for %d to 'Requests done' (current status is %s)" % (jobID, jobStatus),
                )
            stateUpdate = stateServer.setJobStatus(jobID, newJobStatus, "Requests done", "RMS")
            if jobAppStatus and stateUpdate["OK"]:
                stateUpdate = stateServer.setJobApplicationStatus(jobID, jobAppStatus, "RMS")
            if not stateUpdate["OK"]:
                self.log.error(
                    "finalizeRequest: Failed to set job status",
                    "JobID: %d, error: %s" % (jobID, stateUpdate["Message"]),
                )
                return stateUpdate

        return S_OK(newJobStatus)

    @ignoreEncodeWarning
    def getRequestIDsForJobs(self, jobIDs):
        """get the request ids for the supplied jobIDs.

        :param self: self reference
        :param list jobIDs: list of job IDs (integers)
        :return: S_ERROR or S_OK( "Successful": { jobID1: reqID1, jobID2: requID2, ... },
                                  "Failed" : { jobIDn: errMsg, jobIDm: errMsg, ...}  )
        """
        self.log.verbose("getRequestIDsForJobs: attempt to get request(s) for jobs", "(n=%d)" % len(jobIDs))
        res = self._getRPC().getRequestIDsForJobs(jobIDs)
        if not res["OK"]:
            self.log.error(
                "getRequestIDsForJobs: unable to get request(s) for jobs", "%s: %s" % (jobIDs, res["Message"])
            )
            return res

        # Cast the JobIDs back to int
        successful = strToIntDict(res["Value"]["Successful"])
        failed = strToIntDict(res["Value"]["Failed"])

        return S_OK({"Successful": successful, "Failed": failed})

    @ignoreEncodeWarning
    def readRequestsForJobs(self, jobIDs):
        """read requests for jobs

        :param jobIDs: list with jobIDs
        :type jobIDs: python:list
        :return: S_OK( { "Successful" : { jobID1 : Request, ... },
                         "Failed" : { jobIDn : "Fail reason" } } )
        """
        readReqsForJobs = self._getRPC().readRequestsForJobs(jobIDs)
        if not readReqsForJobs["OK"]:
            return readReqsForJobs
        ret = readReqsForJobs["Value"]
        # # create Requests out of JSONs for successful reads
        # Do not forget to cast back str keys to int
        successful = {int(jobID): Request(jsonReq) for jobID, jsonReq in ret["Successful"].items()}
        failed = strToIntDict(ret["Failed"])

        return S_OK({"Successful": successful, "Failed": failed})

    def resetFailedRequest(self, requestID, allR=False):
        """Reset a failed request to "Waiting" status"""

        # # we can safely only peek the request as it is Failed and therefore not owned by an agent
        res = self.peekRequest(requestID)
        if not res["OK"]:
            return res
        req = res["Value"]
        if allR or recoverableRequest(req):
            # Only reset requests that can be recovered
            if req.Status != "Failed":
                gLogger.notice("Reset NotBefore time, was %s" % str(req.NotBefore))
            else:
                for i, op in enumerate(req):
                    op.Error = ""
                    if op.Status == "Failed":
                        printOperation((i, op), onlyFailed=True)
                    for fi in op:
                        if fi.Status == "Failed":
                            fi.Attempt = 1
                            fi.Error = ""
                            fi.Status = "Waiting"
                    if op.Status == "Failed":
                        op.Status = "Waiting"

            # Reset also NotBefore
            req.NotBefore = datetime.datetime.utcnow().replace(microsecond=0)
            return self.putRequest(req)
        return S_OK("Not reset")


# ============= Some useful functions to be shared ===========


output = ""


def prettyPrint(mainItem, key="", offset=0):
    global output
    if key:
        key += ": "
    blanks = offset * " "
    if mainItem and isinstance(mainItem, dict):
        output += "%s%s%s\n" % (blanks, key, "{") if blanks or key else ""
        for key in sorted(mainItem):
            prettyPrint(mainItem[key], key=key, offset=offset)
        output += "%s%s\n" % (blanks, "}") if blanks else ""
    elif mainItem and isinstance(mainItem, list) or isinstance(mainItem, tuple):
        output += "%s%s%s\n" % (blanks, key, "[" if isinstance(mainItem, list) else "(")
        for item in mainItem:
            prettyPrint(item, offset=offset + 2)
        output += "%s%s\n" % (blanks, "]" if isinstance(mainItem, list) else ")")
    elif isinstance(mainItem, six.string_types):
        if "\n" in mainItem:
            prettyPrint(mainItem.strip("\n").split("\n"), offset=offset)
        else:
            output += "%s%s'%s'\n" % (blanks, key, mainItem)
    else:
        output += "%s%s%s\n" % (blanks, key, str(mainItem))
    output = (
        output.replace("[\n%s{" % blanks, "[{")
        .replace("}\n%s]" % blanks, "}]")
        .replace("(\n%s{" % blanks, "({")
        .replace("}\n%s)" % blanks, "})")
        .replace("(\n%s(" % blanks, "((")
        .replace(")\n%s)" % blanks, "))")
        .replace("(\n%s[" % blanks, "[")
        .replace("]\n%s)" % blanks, "]")
    )


def printFTSJobs(request):
    """Prints the FTSJobs associated to a request

    :param request: Request object
    """

    try:
        if request.RequestID:

            # We try first the new FTS3 system

            from DIRAC.DataManagementSystem.Client.FTS3Client import FTS3Client

            fts3Client = FTS3Client()
            res = fts3Client.ping()

            if res["OK"]:
                associatedFTS3Jobs = []
                for op in request:
                    res = fts3Client.getOperationsFromRMSOpID(op.OperationID)
                    if res["OK"]:
                        for fts3Op in res["Value"]:
                            associatedFTS3Jobs.extend(fts3Op.ftsJobs)
                if associatedFTS3Jobs:
                    # Display the direct url and the status
                    gLogger.always(
                        "\n\nFTS3 jobs associated: \n%s"
                        % "\n".join(
                            "%s/fts3/ftsmon/#/job/%s (%s)"
                            % (
                                job.ftsServer.replace(":8446", ":8449"),  # Submission port is 8446, web port is 8449
                                job.ftsGUID,
                                job.status,
                            )
                            for job in associatedFTS3Jobs
                        )
                    )
                return

    # AttributeError can be thrown because the deserialization will not have
    # happened correctly on the new fts3 (CC7 typically), and the error is not
    # properly propagated
    except AttributeError as err:
        gLogger.debug("Could not instantiate FtsClient because of Exception", repr(err))


def printRequest(request, status=None, full=False, verbose=True, terse=False):
    global output

    if full:
        output = ""
        prettyPrint(json.loads(request.toJSON()["Value"]))
        gLogger.always(output)
    else:
        if not status:
            status = request.Status
        gLogger.always(
            "Request name='%s' ID=%s Status='%s'%s%s%s"
            % (
                request.RequestName,
                request.RequestID if hasattr(request, "RequestID") else "(not set yet)",
                request.Status,
                " ('%s' in DB)" % status if status != request.Status else "",
                (" Error='%s'" % request.Error) if request.Error and request.Error.strip() else "",
                (" Job=%s" % request.JobID) if request.JobID else "",
            )
        )
        gLogger.always(
            "Created %s, Updated %s%s"
            % (
                request.CreationTime,
                request.LastUpdate,
                (", NotBefore %s" % request.NotBefore) if request.NotBefore else "",
            )
        )
        if request.OwnerDN:
            gLogger.always("Owner: '%s', Group: %s" % (request.OwnerDN, request.OwnerGroup))
        for indexOperation in enumerate(request):
            op = indexOperation[1]
            if not terse or op.Status == "Failed":
                printOperation(indexOperation, verbose, onlyFailed=terse)
    if not terse:
        printFTSJobs(request)


def printOperation(indexOperation, verbose=True, onlyFailed=False):
    global output
    i, op = indexOperation
    prStr = ""
    if op.SourceSE:
        prStr += "SourceSE: %s" % op.SourceSE
    if op.TargetSE:
        prStr += (" - " if prStr else "") + "TargetSE: %s" % op.TargetSE
    if prStr:
        prStr += " - "
    prStr += "Created %s, Updated %s" % (op.CreationTime, op.LastUpdate)
    if op.Type == "ForwardDISET" and op.Arguments:
        from DIRAC.Core.Utilities import DEncode

        decode, _length = DEncode.decode(op.Arguments)
        if verbose:
            output = ""
            prettyPrint(decode, offset=10)
            prStr += "\n      Arguments:\n" + output.strip("\n")
        else:
            prStr += "\n      Service: %s" % decode[0][0]
    gLogger.always(
        "  [%s] Operation Type='%s' ID=%s Order=%s Status='%s'%s%s"
        % (
            i,
            op.Type,
            op.OperationID if hasattr(op, "OperationID") else "(not set yet)",
            op.Order,
            op.Status,
            (" Error='%s'" % op.Error) if op.Error and op.Error.strip() else "",
            (" Catalog=%s" % op.Catalog) if op.Catalog else "",
        )
    )
    if prStr:
        gLogger.always("      %s" % prStr)
    for indexFile in enumerate(op):
        if not onlyFailed or indexFile[1].Status == "Failed":
            printFile(indexFile)


def printFile(indexFile):
    ind, fi = indexFile
    gLogger.always(
        "    [%02d] ID=%s LFN='%s' Status='%s'%s%s%s"
        % (
            ind + 1,
            fi.FileID if hasattr(fi, "FileID") else "(not set yet)",
            fi.LFN,
            fi.Status,
            (" Checksum='%s'" % fi.Checksum) if fi.Checksum or (fi.Error and "checksum" in fi.Error.lower()) else "",
            (" Error='%s'" % fi.Error) if fi.Error and fi.Error.strip() else "",
            (" Attempts=%d" % fi.Attempt) if fi.Attempt > 1 else "",
        )
    )


def recoverableRequest(request):
    excludedErrors = (
        "File does not exist",
        "No such file or directory",
        "sourceSURL equals to targetSURL",
        "Max attempts limit reached",
        "Max attempts reached",
    )
    operationErrorsOK = ("is banned for", "Failed to perform exists from any catalog")
    for op in request:
        if op.Status == "Failed" and (
            not op.Error or not [errStr for errStr in operationErrorsOK if errStr in op.Error]
        ):
            for fi in op:
                if fi.Status == "Failed":
                    if [errStr for errStr in excludedErrors if errStr in fi.Error]:
                        return False
                    return True
    return True



#!/usr/bin/env python
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
    Script.parseCommandLine(ignoreErrors=False)
    from DIRAC.WorkloadManagementSystem.Client.VirtualMachineCLI import VirtualMachineCLI

    cli = VirtualMachineCLI(vo="enmr.eu")
    cli.cmdloop()


if __name__ == "__main__":
    main()



""" The Script class provides a simple way for users to specify an executable
    or file to run (and is also a simple example of a workflow module).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import io
import re
import stat
import shlex
import distutils.spawn  # pylint: disable=no-name-in-module,no-member,import-error

from DIRAC import gLogger
from DIRAC.Core.Utilities.Subprocess import systemCall

from DIRAC.Workflow.Modules.ModuleBase import ModuleBase


class Script(ModuleBase):
    """Module for running executable"""

    #############################################################################
    def __init__(self, log=None):
        """c'tor"""
        if log is not None:
            self.log = log
        else:
            self.log = gLogger.getSubLogger("Script")
        super(Script, self).__init__(self.log)

        # Set defaults for all workflow parameters here
        self.executable = ""
        self.applicationName = ""
        self.applicationVersion = ""
        self.applicationLog = ""
        self.arguments = ""
        self.workflow_commons = None
        self.step_commons = None

        self.environment = None
        self.callbackFunction = None
        self.bufferLimit = 52428800

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the workflow parameters are resolved here."""
        super(Script, self)._resolveInputVariables()
        super(Script, self)._resolveInputStep()

        self.arguments = self.step_commons.get("arguments", self.arguments)
        if not self.arguments.strip():
            self.arguments = self.workflow_commons.get("arguments", self.arguments)

    #############################################################################

    def _initialize(self):
        """simple checks"""
        if not self.executable:
            raise RuntimeError("No executable defined")

    def _setCommand(self):
        """set the command that will be executed"""
        self.command = self.executable
        if os.path.exists(os.path.basename(self.executable)):
            self.executable = os.path.basename(self.executable)
            if not os.access("%s/%s" % (os.getcwd(), self.executable), 5):
                # doc in https://docs.python.org/2/library/stat.html#stat.S_IRWXU
                os.chmod(
                    "%s/%s" % (os.getcwd(), self.executable),
                    stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
                )
            self.command = "%s/%s" % (os.getcwd(), self.executable)
        elif re.search(".py$", self.executable):
            self.command = "%s %s" % (sys.executable, self.executable)
        elif distutils.spawn.find_executable(self.executable):  # pylint: disable=no-member
            self.command = self.executable

        if self.arguments:
            self.command = "%s %s" % (self.command, self.arguments)

        self.log.info("Command is: %s" % self.command)

    def _executeCommand(self):
        """execute the self.command (uses systemCall)"""
        failed = False

        outputDict = systemCall(
            timeout=0,
            cmdSeq=shlex.split(self.command),
            env=self.environment,
            callbackFunction=self.callbackFunction,
            bufferLimit=self.bufferLimit,
        )
        if not outputDict["OK"]:
            failed = True
            self.log.error("System call execution failed:", "\n" + str(outputDict["Message"]))
        status, stdout, stderr = outputDict["Value"][0:3]
        if status:
            failed = True
            self.log.error("Non-zero status while executing", "%s exited with status %s" % (self.command, status))
        else:
            self.log.info("%s execution completed with status %s" % (self.command, status))

        self.log.verbose(stdout)
        self.log.verbose(stderr)
        if os.path.exists(self.applicationLog):
            self.log.verbose("Removing existing %s" % self.applicationLog)
            os.remove(self.applicationLog)
        with io.open("%s/%s" % (os.getcwd(), self.applicationLog), "wt") as fopen:
            fopen.write(u"<<<<<<<<<< %s Standard Output >>>>>>>>>>\n\n%s " % (self.executable, stdout))
            if stderr:
                fopen.write(u"<<<<<<<<<< %s Standard Error >>>>>>>>>>\n\n%s " % (self.executable, stderr))
        self.log.info("Output written to %s, execution complete." % (self.applicationLog))

        if failed:
            self._exitWithError(status)

    def _exitWithError(self, status):
        """Here because of possible extensions.

        :param str status: the status of the application becomes the status of the workflow,
                           and may be interpreted by JobWrapper (e.g. for rescheduling cases)
        """
        raise RuntimeError(
            "'%s' Exited With Status %s" % (os.path.basename(self.executable).split("_")[0], status), status
        )

    def _finalize(self):
        """simply finalize"""
        applicationString = os.path.basename(self.executable).split("_")[0]
        if self.applicationName and self.applicationName.lower() != "unknown":
            applicationString += " (%s %s)" % (self.applicationName, self.applicationVersion)
        status = "%s successful" % applicationString

        super(Script, self)._finalize(status)



"""
Compile the externals
"""

from __future__ import unicode_literals, absolute_import, division, print_function

import platform
import sys
import os
import re

__RCSID__ = "$Id$"

# We need to patch python platform module. It does a string comparison for the libc versions.
# it fails when going from 2.9 to 2.10,
# the fix converts the version to a tuple and attempts a numeric comparison

_libc_search = re.compile(b"(__libc_init)" b"|" b"(GLIBC_([0-9.]+))" b"|" b"(libc(_\\w+)?\\.so(?:\\.(\\d[0-9.]*))?)")


def libc_ver(executable=sys.executable, lib="", version="", chunksize=2048):
    """Tries to determine the libc version that the file executable
    (which defaults to the Python interpreter) is linked against.

    Returns a tuple of strings (lib,version) which default to the
    given parameters in case the lookup fails.

    Note that the function has intimate knowledge of how different
    libc versions add symbols to the executable and thus is probably
    only useable for executables compiled using gcc.

    The file is read and scanned in chunks of chunksize bytes.

    """
    with open(executable, "rb") as f:
        binary = f.read(chunksize)
        pos = 0
        version = [0, 0, 0]
        while True:
            m = _libc_search.search(binary, pos)
            if not m:
                binary = f.read(chunksize)
                if not binary:
                    break
                pos = 0
                continue
            libcinit, glibc, glibcversion, so, threads, soversion = m.groups()
            if libcinit and not lib:
                lib = b"libc"
            elif glibc:
                glibcversion_parts = glibcversion.split(b".")
                for i in range(len(glibcversion_parts)):
                    try:
                        glibcversion_parts[i] = int(glibcversion_parts[i])
                    except ValueError:
                        glibcversion_parts[i] = 0
                if libcinit and not lib:
                    lib = b"libc"
                elif glibc:
                    if lib != b"glibc":
                        lib = b"glibc"
                        version = glibcversion_parts
                    elif glibcversion_parts > version:
                        version = glibcversion_parts
            elif so:
                if lib != b"glibc":
                    lib = b"libc"
                    version = max(version, [int(soversion)]).pop()
                    if threads and version[-len(threads) :] != threads:
                        version = version + threads
            pos = m.end()
    return lib.decode(), ".".join(map(str, version))


# ## Command line interface


def getPlatformString():
    # Modified to return our desired platform string, R. Graciani
    platformTuple = (platform.system(), platform.machine())
    if platformTuple[0] == "Linux":
        try:
            import subprocess

            sp = subprocess.Popen(["/sbin/ldconfig", "--print-cache"], stdout=subprocess.PIPE, universal_newlines=True)
            spStdout = sp.stdout
        except Exception:
            sp = None
            spStdout = os.popen("/sbin/ldconfig --print-cache", "r")
        ldre = re.compile(r".*=> (.*/libc\.so\..*$)")
        libs = []
        for line in spStdout.readlines():
            reM = ldre.match(line)
            if reM:
                libs.append(reM.groups()[0])
        if sp:
            if "terminate" in dir(sp):
                sp.terminate()
                sp.wait()

        if not libs:
            # get version of higher libc installed
            if platform.machine().find("64") != -1:
                lib = "/lib64"
            else:
                lib = "/lib"
            for libFile in os.listdir(lib):
                if libFile.find("libc-") == 0 or libFile.find("libc.so") == 0:
                    libs.append(os.path.join(lib, libFile))
        newest_lib = [0, 0, 0]
        for lib in libs:
            lib_parts = libc_ver(lib)[1].split(".")
            for i in range(len(lib_parts)):
                try:
                    lib_parts[i] = int(lib_parts[i])
                except ValueError:
                    lib_parts[i] = 0
                    # print "non integer version numbers"
            if lib_parts > newest_lib:
                newest_lib = lib_parts

        platformTuple += ("glibc-" + ".".join(map(str, newest_lib)),)
    elif platformTuple[0] == "Darwin":
        platformTuple += (".".join(platform.mac_ver()[0].split(".")[:2]),)
    elif platformTuple[0] == "Windows":
        platformTuple += (platform.win32_ver()[0],)
    else:
        platformTuple += platform.release()

    platformString = "%s_%s_%s" % platformTuple

    return platformString


_gPlatform = None
_gPlatformTuple = None


def getPlatform():
    global _gPlatform, _gPlatformTuple

    if _gPlatform is None:
        _gPlatform = getPlatformString()
        _gPlatformTuple = tuple(_gPlatform.split("_"))

    return _gPlatform


def getPlatformTuple():
    getPlatform()
    return _gPlatformTuple



""" This contains unit tests to make sure that the migration between PyGSI and M2Crypto is as smooth as possible
    The test covers only the method exposed by the PyGSI version for the time being.

    We are not testing:

    * generateProxyRequest -> boils down to testing X509Request
    * setCertificate -> not used anywhere

    We are skipping:
    * getPublicKey -> no way to test that really
    * getSerialNumber -> buggy in PyGSI

"""

# redefined-outer-name is needed because we keep passing get_X509Certificate_class as param
# pylint: disable=redefined-outer-name

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from DIRAC.Core.Security.test.x509TestUtilities import (
    deimportDIRAC,
    CERTS,
    CERTCONTENTS,
    getCertOption,
    HOSTCERT,
    VOMSPROXY,
    VOMS_PROXY_ATTR,
)

from pytest import mark, fixture, skip

parametrize = mark.parametrize

X509CERTTYPES = ("M2_X509Certificate",)

# This fixture will return a X509Certificate class
# https://docs.pytest.org/en/latest/fixture.html#automatic-grouping-of-tests-by-fixture-instances


@fixture(scope="function", params=X509CERTTYPES)
def get_X509Certificate_class(request):
    """Fixture to return the X509Certificate class.
    It also 'de-import' DIRAC before and after
    """
    # Clean before
    deimportDIRAC()

    x509Class = request.param

    if x509Class == "M2_X509Certificate":
        from DIRAC.Core.Security.m2crypto.X509Certificate import X509Certificate
    else:
        raise NotImplementedError()

    yield X509Certificate

    # Clean after
    deimportDIRAC()


@parametrize("cert_file", CERTS)
def test_executeOnlyIfCertLoaded(cert_file, get_X509Certificate_class):
    """ " Tests whether the executeOnlyIfCertLoaded decorator works"""
    x509Cert = get_X509Certificate_class()
    # Since we did not load the certificate, we should get S_ERROR
    res = x509Cert.getNotAfterDate()
    from DIRAC.Core.Utilities.DErrno import ENOCERT

    assert res["Errno"] == ENOCERT

    # Now load it
    x509Cert.load(cert_file)
    res = x509Cert.getNotAfterDate()
    assert res["OK"]


@parametrize("cert_file", CERTS)
def test_load(cert_file, get_X509Certificate_class):
    """ " Just load a certificate"""
    x509Cert = get_X509Certificate_class()
    res = x509Cert.load(cert_file)
    assert res["OK"]


def test_load_non_existing_file(get_X509Certificate_class):
    """ " Just load a non existing file and non pem formated string"""
    x509Cert = get_X509Certificate_class()
    res = x509Cert.load("/tmp/nonexistingFile.pem")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import ECERTREAD

    assert res["Errno"] == ECERTREAD


@parametrize("cert_file", CERTS)
def test_loadFromFile(cert_file, get_X509Certificate_class):
    """ " Just load a certificate"""
    x509Cert = get_X509Certificate_class()
    res = x509Cert.loadFromFile(cert_file)
    assert res["OK"]


def test_loadFromFile_non_existing_file(get_X509Certificate_class):
    """ " Just load a non existing file"""
    x509Cert = get_X509Certificate_class()
    res = x509Cert.loadFromFile("/tmp/nonexistingFile.pem")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import EOF

    assert res["Errno"] == EOF


# pylint: disable=unused-argument
@parametrize("cert_content_type", CERTCONTENTS)
def test_loadFromString(cert_content_type, get_X509Certificate_class, indirect=("hostcertcontent", "usercertcontent")):
    """ " Just load a certificate from PEM string
    :param cert_content_type: either HOSTCERTCONTENT or USERCERTCONTENT
    :param indirect: pytest trick,
          see https://docs.pytest.org/en/latest/example/parametrize.html#apply-indirect-on-particular-arguments
    """
    x509Cert = get_X509Certificate_class()
    res = x509Cert.loadFromString(CERTCONTENTS[cert_content_type])
    assert res["OK"], res


def test_loadFromString_non_pem(get_X509Certificate_class):
    """ " Just load a non pem formated string"""
    x509Cert = get_X509Certificate_class()
    res = x509Cert.loadFromString("THIS IS NOT PEM DATA")
    assert not res["OK"]

    from DIRAC.Core.Utilities.DErrno import ECERTREAD

    assert res["Errno"] == ECERTREAD


# TODO: have a non valid certificate to try
@parametrize("cert_file", CERTS)
def test_hasExpired(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check it has not expired"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.hasExpired()
    assert res["OK"]
    assert not res["Value"]


@parametrize("cert_file", CERTS)
def test_getNotAfterDate(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check its expiration date"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getNotAfterDate()

    assert res["OK"]
    # We expect getNotAfterDate to return a datetime
    assert res["Value"].date() == getCertOption(cert_file, "endDate")


@parametrize("cert_file", CERTS)
def test_getNotBeforeDate(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check its start validity date"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getNotBeforeDate()

    assert res["OK"]
    # We expect getNotBeforeDate to return a datetime
    assert res["Value"].date() == getCertOption(cert_file, "startDate")


@parametrize("cert_file", CERTS)
def test_getSubjectDN(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check its subject"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getSubjectDN()

    assert res["OK"]
    assert res["Value"] == getCertOption(cert_file, "subjectDN")


@parametrize("cert_file", CERTS)
def test_getIssuerDN(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check its issuer"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getIssuerDN()

    assert res["OK"]
    assert res["Value"] == getCertOption(cert_file, "issuerDN")


# # TODO: this method seems not to be used anyway
# @parametrize('cert_file', CERTS)
# def test_getSubjectNameObject(cert_file, get_X509Certificate_class):
#   """" Load a valid certificate and check its subject object. """
#   x509Cert = get_X509Certificate_class()
#   x509Cert.load(cert_file)

#   res = x509Cert.getSubjectNameObject()

#   assert res['OK']
#   # We cannot compare the objects themselves because it is a different object...
#   expectedValue = getCertOption(cert_file, 'subjectDN')
#   try:
#     # This works in the case of pyGSI
#     returnedValue = res['Value'].one_line()
#   except AttributeError:
#     # This works in the case of M2Crypto
#     returnedValue = str(res['Value'])

#   assert returnedValue == expectedValue


# # TODO: this method seems not to be used anyway
# @parametrize('cert_file', CERTS)
# def test_getIssuerNameObject(cert_file, get_X509Certificate_class):
#   """" Load a valid certificate and check its subject object. """
#   x509Cert = get_X509Certificate_class()
#   x509Cert.load(cert_file)

#   res = x509Cert.getIssuerNameObject()

#   assert res['OK']
#   # We cannot compare the objects themselves because it is a different object...
#   expectedValue = getCertOption(cert_file, 'issuerDN')
#   try:
#     # This works in the case of pyGSI
#     returnedValue = res['Value'].one_line()
#   except AttributeError:
#     # This works in the case of M2Crypto
#     returnedValue = str(res['Value'])

#   assert returnedValue == expectedValue


@mark.skip(reason="no way of currently testing this")
@parametrize("cert_file", CERTS)
def test_getPublicKey(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and verify its public key (for m2crypto only)"""
    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getPublicKey()
    assert res["OK"]

    if "m2crypto" in get_X509Certificate_class.__module__:
        print(x509Cert.verify(res["Value"]))


@parametrize("cert_file", CERTS)
def test_getSerialNumber(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check its public key"""

    x509Cert = get_X509Certificate_class()

    x509Cert.load(cert_file)

    res = x509Cert.getSerialNumber()

    assert res["OK"]
    assert res["Value"] == getCertOption(cert_file, "serial")


@parametrize("cert_file", CERTS)
def test_getDIRACGroup_on_cert(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check if there is a dirac group. It should not"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    # ignoreDefault is used such that there is no attempt to look for group in the CS
    res = x509Cert.getDIRACGroup(ignoreDefault=True)

    assert res["OK"]
    assert res["Value"] is False


@parametrize("cert_file", CERTS)
def test_hasVOMSExtensions_on_cert(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check if it has VOMS extensions. It should not"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    # ignoreDefault is used such that there is no attempt to look for group in the CS
    res = x509Cert.hasVOMSExtensions()

    assert res["OK"]
    assert res["Value"] is False


@parametrize("cert_file", CERTS)
def test_getVOMSData_on_cert(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and load the (non existing VOMS data)"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getVOMSData()
    from DIRAC.Core.Utilities.DErrno import EVOMS

    assert not res["OK"]
    assert res["Errno"] == EVOMS


@parametrize("cert_file", CERTS)
def test_getRemainingSecs_on_cert(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check the output is a positive integer"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getRemainingSecs()

    assert res["OK"]
    assert isinstance(res["Value"], int) and res["Value"] > 0


@parametrize("cert_file", CERTS)
def test_getExtensions_on_cert(cert_file, get_X509Certificate_class):
    """ " Load a valid certificate and check the output is a positive integer"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(cert_file)

    res = x509Cert.getExtensions()

    assert res["OK"]

    extensionDict = dict(extTuple for extTuple in res["Value"])

    assert sorted(extensionDict) == sorted(getCertOption(cert_file, "availableExtensions"))

    # Test a few of them
    for ext in ("basicConstraints", "extendedKeyUsage"):
        assert extensionDict[ext] == getCertOption(cert_file, ext)

    # Valid only for Host certificate:
    if cert_file == HOSTCERT:
        assert extensionDict["subjectAltName"] == getCertOption(cert_file, "subjectAltName")


###########################################################################
# Temporary. For the time being, we need a real proxy !


def test_getVOMSData(get_X509Certificate_class):
    """ " Load a valid certificate and check the output is a positive integer"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(VOMSPROXY)

    res = x509Cert.getVOMSData()
    assert res["OK"]
    assert res["Value"] == VOMS_PROXY_ATTR


def test_hasVOMSExtensions(get_X509Certificate_class):
    """ " Load a certificate generated with voms-proxy-fake and check hasVOMSExtension is True"""

    x509Cert = get_X509Certificate_class()
    x509Cert.load(VOMSPROXY)
    res = x509Cert.hasVOMSExtensions()
    assert res["OK"]
    assert res["Value"]



""" X509CRL is a class for managing X509CRL
This class is used to manage the revoked certificates....
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import stat
import os
import tempfile
import re
import datetime

import M2Crypto
import six
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno

# pylint: disable=broad-except


class X509CRL(object):
    def __init__(self, cert=None):
        self.__pemData = b""

        if cert:
            self.__loadedCert = True
            self.__revokedCert = cert
        else:
            self.__loadedCert = False

    @classmethod
    def instanceFromFile(cls, crlLocation):
        """Instance a X509CRL from a file"""
        crl = cls()
        result = crl.loadCRLFromFile(crlLocation)
        if not result["OK"]:
            return result
        return S_OK(crl)

    def loadCRLFromFile(self, crlLocation):
        """
        Load a x509CRL certificate from a pem file
        Return : S_OK / S_ERROR
        """
        self.__loadedCert = False
        try:
            self.__revokedCert = M2Crypto.X509.load_crl(crlLocation)
        except Exception as e:
            return S_ERROR(DErrno.ECERTREAD, "%s" % repr(e).replace(",)", ")"))
        self.__loadedCert = True
        with open(crlLocation, "rb") as crlFile:
            pemData = crlFile.read()
        self.__pemData = pemData
        return S_OK()

    def __bytes__(self):
        if not self.__loadedCert:
            return b"No certificate loaded"
        return self.__pemData

    if six.PY2:

        def __str__(self):
            return self.__bytes__()

    else:

        def __str__(self):
            return bytes(self).decode()

    def dumpAllToString(self):
        """
        Dump all to string
        """
        if not self.__loadedCert:
            return S_ERROR(DErrno.ECERTREAD, "No certificate loaded")
        return S_OK(self.__pemData)

    def dumpAllToFile(self, filename=False):
        """
        Dump all to file. If no filename specified a temporal one will be created
        """
        if not self.__loadedCert:
            return S_ERROR("No certificate loaded")
        try:
            if not filename:
                fd, filename = tempfile.mkstemp()
                os.close(fd)
            with open(filename, "wb") as fd:
                fd.write(self.__pemData)
        except Exception as e:
            return S_ERROR(DErrno.EWF, "%s: %s" % (filename, repr(e).replace(",)", ")")))
        try:
            os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)
        except Exception as e:
            return S_ERROR(DErrno.ESPF, "%s: %s" % (filename, repr(e).replace(",)", ")")))
        return S_OK(filename)

    def hasExpired(self):
        if not self.__loadedCert:
            return S_ERROR("No certificate loaded")
        # XXX It should be done better, for now M2Crypto doesn't offer access to fields like Next Update
        txt = self.__revokedCert.as_text()
        pattern = r"Next Update: (?P<nextUpdate>.*)\n"
        dateStr = re.search(pattern.encode("utf-8"), txt).group("nextUpdate")
        nextUpdate = datetime.datetime.strptime(dateStr, "%b %d %H:%M:%S %Y GMT")
        return S_OK(datetime.datetime.now() > nextUpdate)

    def getIssuer(self):
        if not self.__loadedCert:
            return S_ERROR("No certificate loaded")
        # XXX It should be done better, for now M2Crypto doesn't offer access to fields like Issuer
        txt = self.__revokedCert.as_text()
        pattern = r"Issuer: (?P<issuer>.*)\n"
        return S_OK(re.search(pattern.encode("utf-8"), txt).group("issuer"))

    def __repr__(self):
        repStr = "<X509CRL"
        if self.__loadedCert:
            repStr += ""  # self.__revokedCert.get_issuer().one_line()  # Why issuer?! XXX
        repStr += ">"
        return repStr



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-delete-user
# Author :  Adrian Casajus
########################################################################
"""
Remove User from Configuration

Example:
  $ dirac-admin-delete-user vhamar
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["User:     User name"])
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    choice = six.moves.input("Are you sure you want to delete user/s %s? yes/no [no]: " % ", ".join(args))
    choice = choice.lower()
    if choice not in ("yes", "y"):
        print("Delete aborted")
        DIRACExit(0)

    for user in args:
        if not diracAdmin.csDeleteUser(user):
            errorList.append(("delete user", "Cannot delete user %s" % user))
            exitCode = 255

    if not exitCode:
        result = diracAdmin.csCommitChanges()
        if not result["OK"]:
            errorList.append(("commit", result["Message"]))
            exitCode = 255

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



"""
Test LogLevels module
"""

__RCSID__ = "$Id$"

import logging
import pytest

from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels


@pytest.mark.parametrize(
    "logLevel, value",
    [
        ("debug", logging.DEBUG),
        ("verbose", 15),
        ("info", 20),
        ("warn", logging.WARN),
        ("notice", 35),
        ("error", logging.ERROR),
        ("always", 45),
        ("fatal", logging.CRITICAL),
        ("unknown", None),
    ],
)
def test_getLevelValue(logLevel, value):
    """
    Test getLevelValue
    """
    assert LogLevels.getLevelValue(logLevel) == value


@pytest.mark.parametrize(
    "value, logLevel",
    [
        (logging.DEBUG, "debug"),
        (15, "verbose"),
        (20, "info"),
        (logging.WARN, "warn"),
        (35, "notice"),
        (logging.ERROR, "error"),
        (45, "always"),
        (logging.CRITICAL, "fatal"),
        (-1, None),
        ("error", None),
    ],
)
def test_getLevel(value, logLevel):
    """
    Test getLevel
    """
    if logLevel:
        logLevel = logLevel.upper()
    assert LogLevels.getLevel(value) == logLevel


def test_getLevelNames():
    """
    Test getLevelNames
    """
    levels = ["DEBUG", "VERBOSE", "INFO", "WARN", "NOTICE", "ERROR", "ALWAYS", "FATAL"]
    assert sorted(LogLevels.getLevelNames()) == sorted(levels)
